/*  RAxML-VI-HPC (version 2.2) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright August 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  Alexandros.Stamatakis@epfl.ch
 *
 *  When publishing work that is based on the results from RAxML-VI-HPC please cite:
 *
 *  Alexandros Stamatakis:"RAxML-VI-HPC: maximum likelihood-based phylogenetic analyses with thousands of taxa and mixed models". 
 *  Bioinformatics 2006; doi: 10.1093/bioinformatics/btl446
 */


#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h> 
#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "axml.h"

extern char AAfreq_noselect_FileName[1024];

double sum_pi_jc[20];  /* sum of aa freq of 4 PCA classes that are under selection */
double protModelFreq[20];

extern int optimizeRatesInvocations;
extern int optimizeRateCategoryInvocations;
extern int optimizeAlphaInvocations;
extern int optimizeTTRatioInvocations;

extern int processID;
extern char userModelsFileName[1024];

extern boolean (*newview)           (tree *, nodeptr);
extern double  (*evaluate)          (tree *, nodeptr);   
extern double  (*makenewz)          (tree *, nodeptr, nodeptr, double, int);
extern boolean (*myNewview)         (tree *, int, nodeptr);
extern double  (*myEvaluate)        (tree *, int, nodeptr);   


extern double  (*evaluatePartial)   (tree *, nodeptr, int, double);

extern double  (*evaluatePartition) (tree *, nodeptr, int);
extern boolean (*newviewPartition)  (tree *, nodeptr, int);


void makeVal(char code, double *val)
{
  double i_value;
  int j;

  if(code < 0 || code > 22)
    {
      printf("FATAL ERROR in makeVal() \n");
      exit(-1);
    }

  if(code < 22)
    i_value = 0.0;
  else
    i_value = 1.0;	 

  for(j = 0; j < 20; j++)
    val[j] = i_value;
  
  if(code < 20)
    val[code] = 1.0;
  else
    {
      if(code == 20)	  
	val[2] = val[3] = 1.0;
      if(code == 21)
	val[5] = val[6] = 1.0;	  		   
    }
}


void smoothFreqs(const int n, double *pfreqs, double *dst)
{
  int 
    countScale = 0, 
    l,
    loopCounter = 0;

  for(l = 0; l < n; l++)
    if(pfreqs[l] < FREQ_MIN)
      countScale++;

  if(countScale > 0)
    {	     
      while(countScale > 0)
	{
	  double correction = 0.0;
	  double factor = 1.0;
	  
	  for(l = 0; l < n; l++)
	    {
	      if(pfreqs[l] == 0.0)		  
		correction += FREQ_MIN;		   		  
	      else
		if(pfreqs[l] < FREQ_MIN)		    
		  {
		    correction += (FREQ_MIN - pfreqs[l]);
		    factor -= (FREQ_MIN - pfreqs[l]);
		  }
	    }		      	    	    
	  
	  countScale = 0;
	  
	  for(l = 0; l < n; l++)
	    {		    
	      if(pfreqs[l] >= FREQ_MIN)		      
		pfreqs[l] = pfreqs[l] - (pfreqs[l] * correction * factor);	
	      else
		pfreqs[l] = FREQ_MIN;
	      
	      if(pfreqs[l] < FREQ_MIN)
		countScale++;
	    }
	  assert(loopCounter < 100);
	  loopCounter++;
	}		    
    }

  for(l = 0; l < n; l++)
    dst[l] = pfreqs[l];

  /*  
  if(partitionData->nonGTR)
    {
      int k;

      assert(partitionData->dataType == SECONDARY_DATA_7 || partitionData->dataType == SECONDARY_DATA_6);
       
      for(l = 0; l < n; l++)
	{
	  int count = 1;	
	  
	  for(k = 0; k < n; k++)
	    {
	      if(k != l && partitionData->frequencyGrouping[l] == partitionData->frequencyGrouping[k])
		{
		  count++;
		  dst[l] += pfreqs[k];
		}
	    }
	  dst[l] /= ((double)count);
	}            
     }
  */

}


void baseFrequenciesGTR(rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef)
{
  double  sum, suma, sumc, sumg, sumt, wj, fa, fc, fg, ft, freqa, freqc, freqg, freqt, pfreqs[20], sumf[20], val[20], temp[20], acc;
  int     i, j, k, l, code, y;
  char  *yptr;

  if(!adef->useMultipleModel)
    {
      if(adef->model == M_PROTCAT || adef->model == M_PROTGAMMA)
	{	 	  	 	  
	  for(l = 0; l < 20; l++)	    
	    pfreqs[l] = 0.05;	     	    

	  for (k = 1; k <= 8; k++) 
	    {	     	   	    	      			
	      for(l = 0; l < 20; l++)
		sumf[l] = 0.0;

	      for (i = 0; i < rdta->numsp; i++) 
		{		 
		  yptr = rdta->y[i];
		  for(j = 0; j < cdta->endsite; j++) 
		    {		      		     		    		     		      
		      makeVal(yptr[j], val);		     				     				     
		      for(l = 0; l < 20; l++)						
			temp[l] = pfreqs[l] * val[l];
			  						  
		      acc = 0;
		      for(l = 0; l < 20; l++)
			{
			  if(temp[l] != 0.0)
			    acc += temp[l];
			}		      		      
		      
		      wj = cdta->aliaswgt[j] / acc;
		      
		      for(l = 0; l < 20; l++)
			{
			  if(temp[l] != 0.0)
			    {
			      sumf[l] += wj * temp[l];			     				   			     
			    }
			}
		    }
		}	    
	  
	      acc = 0;
	      for(l = 0; l < 20; l++)
		acc += sumf[l];
	      for(l = 0; l < 20; l++)
		pfreqs[l] = sumf[l] / acc;	     
	    }

	  smoothFreqs(20, pfreqs, tr->frequencies);
	  	 
	  for(l = 0; l < 20; l++) 
	    {	    
	      tr->frequencies[l] = pfreqs[l];
	    }	     
    }
  else
    {
      freqa = 0.25;
      freqc = 0.25;
      freqg = 0.25;
      freqt = 0.25;
      for (k = 1; k <= 8; k++) 
	{
	  suma = 0.0;
	  sumc = 0.0;
	  sumg = 0.0;
	  sumt = 0.0;
	      for (i = 0; i < rdta->numsp; i++) 
		{
		  yptr = rdta->y[i];
		  for (j = 0; j < cdta->endsite; j++) 
		    {
		      code = *yptr++;
		      /*if(code == 0)
			printf("KOT 0\n");*/
		      if(code > 0)
			{
			  fa = freqa * ( code       & 1);
			  fc = freqc * ((code >> 1) & 1);
			  fg = freqg * ((code >> 2) & 1);
			  ft = freqt * ((code >> 3) & 1);
			  wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
			  suma += wj * fa;
			  sumc += wj * fc;
			  sumg += wj * fg;
			  sumt += wj * ft;
			}
		    }
		}
	      sum = suma + sumc + sumg + sumt;
	      freqa = suma / sum;
	      freqc = sumc / sum;
	      freqg = sumg / sum;
	      freqt = sumt / sum;
	    }
      
	  tr->frequencies[0] = freqa;
	  tr->frequencies[1] = freqc;
	  tr->frequencies[2] = freqg;
	  tr->frequencies[3] = freqt;
	}
    }
  else
    {
      int model;
      int lower;
      int upper;

      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  lower = tr->modelIndices[model][0];
	  upper = tr->modelIndices[model][1];	  	 

	  if(adef->model == M_PROTCAT || adef->model == M_PROTGAMMA)
	    {	 	  	 	  
	      for(l = 0; l < 20; l++)	    
		pfreqs[l] = 0.05;	     	    
	      
	      for (k = 1; k <= 8; k++) 
		{	     	   	    	      			
		  for(l = 0; l < 20; l++)
		    sumf[l] = 0.0;
		  
		  for (i = 0; i < rdta->numsp; i++) 
		    {		 
		      yptr = rdta->y[i];
		      for(j = lower; j < upper; j++) 
			{		      		     		    		     		      
			  makeVal(yptr[j], val);		     				     				     
			  for(l = 0; l < 20; l++)						
			    temp[l] = pfreqs[l] * val[l];
			  
			  acc = 0;
			  for(l = 0; l < 20; l++)
			    {
			      if(temp[l] != 0.0)
				acc += temp[l];
			    }
			  
			  
			  
			  wj = cdta->aliaswgt[j] / acc;
			  
			  for(l = 0; l < 20; l++)
			    {
			      if(temp[l] != 0.0)
				{
				  sumf[l] += wj * temp[l];			     				   			     
				}
			    }
			}
		    }	    
		  
		  acc = 0;
		  for(l = 0; l < 20; l++)
		    acc += sumf[l];
		  for(l = 0; l < 20; l++)
		    pfreqs[l] = sumf[l] / acc;	     
		}

	      smoothFreqs(20, pfreqs, tr->frequencies);
	  	 
	      for(l = 0; l < 20; l++) 
		{	    
		  tr->frequencies[model * 20 + l] = pfreqs[l];		 
		}	     
	    }
	  else
	    {

	      freqa = 0.25;
	      freqc = 0.25;
	      freqg = 0.25;
	      freqt = 0.25;
	      for (k = 1; k <= 8; k++) 
		{
		  suma = 0.0;
		  sumc = 0.0;
		  sumg = 0.0;
		  sumt = 0.0;
		  for (i = 0; i < rdta->numsp; i++) 
		    {
		      yptr = &(rdta->y[i][lower]);
		      for (j = lower; j < upper; j++) 
			{
			  code = *yptr++;
			  if(code > 0)
			    {
			      fa = freqa * ( code       & 1);
			      fc = freqc * ((code >> 1) & 1);
			      fg = freqg * ((code >> 2) & 1);
			      ft = freqt * ((code >> 3) & 1);
			      wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
			      suma += wj * fa;
			      sumc += wj * fc;
			      sumg += wj * fg;
			      sumt += wj * ft;
			    }
			}
		    }
		  sum = suma + sumc + sumg + sumt;
		  freqa = suma / sum;
		  freqc = sumc / sum;
		  freqg = sumg / sum;
		  freqt = sumt / sum;
		}
	      
	      tr->frequencies[model * 4] = freqa;
	      tr->frequencies[model * 4 + 1] = freqc;
	      tr->frequencies[model * 4 + 2] = freqg;
	      tr->frequencies[model * 4 + 3] = freqt;
	    }
	  
	}      
    }
  return;
}



void readExternalRateMatrix (double *initialRates, double f[20], char *matrixFileName)
//void readExternalRateMatrix (char *matrixFileName)
{
  FILE *myFile;
  int i, j, count, r;
  double aaMut[20][20], a[20][20];
  char *pch, buf[256];
  
  myFile = fopen(matrixFileName,"r");

  if (!myFile) {
    if (processID == 0)
      printf("Could not open the external rate matrix file: %s\n", matrixFileName);
    errorExit(-1);
  }
    
  /* read rate matrix */
  i = 1; j = 0; /* the diagnal values not provideed in the exchange matrix, so the matrix start at i=1*/   
  for (count = 1; count < 20; count ++) { 
    fgets (buf, 256, myFile);
    pch = strtok(buf, " ");
    while (pch != NULL) {
      if (*pch == '\n') break;
      aaMut[i][j++] = atof(pch);
      pch = strtok(NULL, " ");
    }
    i ++; j = 0;
  }
      
  /*  get aa freq */
  fgets (buf, 256, myFile);   // skip the blank line
  fgets (buf, 256, myFile);  
  i = 0;
  pch = strtok(buf, " ,\t");
  while (pch != NULL){
    if (*pch == '\n') break;
    f[i++] = atof(pch);
    pch = strtok (NULL, " ,\t");
  }

  /* from triangular exchange matrix to square exchange  matrix */
  for (i=0; i<20; i++) {
    for (j=0; j<=i; j++) 
      a[i][j] = aaMut[i][j];
    
    for (j=i+1; j<20; j++)
      a[i][j] = aaMut[j][i];  
  }

  r = 0;
  for(i = 0; i < 19; i++)
    {      
      for(j = i + 1; j < 20; j++) 
	initialRates[r++] = a[i][j];
    }

  fclose (myFile);
  
}


int initProtMat(tree *tr, int model, double f[20], analdef *adef, int proteinMatrix)
{ 
  double q[20][20];
  double daa[400], max, temp;
  int i, j, r;
  double *initialRates = &(tr->initialRates[model * tr->numberOfRates]);


  switch(proteinMatrix)
    {
    case F81:
      {

	double aaMut[20][20], a[20][20];

aaMut[1][0]=1.0; 
aaMut[2][0]=1.0; aaMut[2][1]=1.0; 
aaMut[3][0]=1.0; aaMut[3][1]=1.0; aaMut[3][2]=1.0; 
aaMut[4][0]=1.0; aaMut[4][1]=1.0; aaMut[4][2]=1.0; aaMut[4][3]=1.0; 
aaMut[5][0]=1.0; aaMut[5][1]=1.0; aaMut[5][2]=1.0; aaMut[5][3]=1.0; aaMut[5][4]=1.0; 
aaMut[6][0]=1.0; aaMut[6][1]=1.0; aaMut[6][2]=1.0; aaMut[6][3]=1.0; aaMut[6][4]=1.0; aaMut[6][5]=1.0; 
aaMut[7][0]=1.0; aaMut[7][1]=1.0; aaMut[7][2]=1.0; aaMut[7][3]=1.0; aaMut[7][4]=1.0; aaMut[7][5]=1.0; aaMut[7][6]=1.0; 
aaMut[8][0]=1.0; aaMut[8][1]=1.0; aaMut[8][2]=1.0; aaMut[8][3]=1.0; aaMut[8][4]=1.0; aaMut[8][5]=1.0; aaMut[8][6]=1.0; aaMut[8][7]=1.0; 
aaMut[9][0]=1.0; aaMut[9][1]=1.0; aaMut[9][2]=1.0; aaMut[9][3]=1.0; aaMut[9][4]=1.0; aaMut[9][5]=1.0; aaMut[9][6]=1.0; aaMut[9][7]=1.0; aaMut[9][8]=1.0; 
aaMut[10][0]=1.0; aaMut[10][1]=1.0; aaMut[10][2]=1.0; aaMut[10][3]=1.0; aaMut[10][4]=1.0; aaMut[10][5]=1.0; aaMut[10][6]=1.0; aaMut[10][7]=1.0; aaMut[10][8]=1.0; aaMut[10][9]=1.0; 
aaMut[11][0]=1.0; aaMut[11][1]=1.0; aaMut[11][2]=1.0; aaMut[11][3]=1.0; aaMut[11][4]=1.0; aaMut[11][5]=1.0; aaMut[11][6]=1.0; aaMut[11][7]=1.0; aaMut[11][8]=1.0; aaMut[11][9]=1.0; aaMut[11][10]=1.0; 
aaMut[12][0]=1.0; aaMut[12][1]=1.0; aaMut[12][2]=1.0; aaMut[12][3]=1.0; aaMut[12][4]=1.0; aaMut[12][5]=1.0; aaMut[12][6]=1.0; aaMut[12][7]=1.0; aaMut[12][8]=1.0; aaMut[12][9]=1.0; aaMut[12][10]=1.0; aaMut[12][11]=1.0; 
aaMut[13][0]=1.0; aaMut[13][1]=1.0; aaMut[13][2]=1.0; aaMut[13][3]=1.0; aaMut[13][4]=1.0; aaMut[13][5]=1.0; aaMut[13][6]=1.0; aaMut[13][7]=1.0; aaMut[13][8]=1.0; aaMut[13][9]=1.0; aaMut[13][10]=1.0; aaMut[13][11]=1.0; aaMut[13][12]=1.0; 
aaMut[14][0]=1.0; aaMut[14][1]=1.0; aaMut[14][2]=1.0; aaMut[14][3]=1.0; aaMut[14][4]=1.0; aaMut[14][5]=1.0; aaMut[14][6]=1.0; aaMut[14][7]=1.0; aaMut[14][8]=1.0; aaMut[14][9]=1.0; aaMut[14][10]=1.0; aaMut[14][11]=1.0; aaMut[14][12]=1.0; aaMut[14][13]=1.0; 
aaMut[15][0]=1.0; aaMut[15][1]=1.0; aaMut[15][2]=1.0; aaMut[15][3]=1.0; aaMut[15][4]=1.0; aaMut[15][5]=1.0; aaMut[15][6]=1.0; aaMut[15][7]=1.0; aaMut[15][8]=1.0; aaMut[15][9]=1.0; aaMut[15][10]=1.0; aaMut[15][11]=1.0; aaMut[15][12]=1.0; aaMut[15][13]=1.0; aaMut[15][14]=1.0; 
aaMut[16][0]=1.0; aaMut[16][1]=1.0; aaMut[16][2]=1.0; aaMut[16][3]=1.0; aaMut[16][4]=1.0; aaMut[16][5]=1.0; aaMut[16][6]=1.0; aaMut[16][7]=1.0; aaMut[16][8]=1.0; aaMut[16][9]=1.0; aaMut[16][10]=1.0; aaMut[16][11]=1.0; aaMut[16][12]=1.0; aaMut[16][13]=1.0; aaMut[16][14]=1.0; aaMut[16][15]=1.0; 
aaMut[17][0]=1.0; aaMut[17][1]=1.0; aaMut[17][2]=1.0; aaMut[17][3]=1.0; aaMut[17][4]=1.0; aaMut[17][5]=1.0; aaMut[17][6]=1.0; aaMut[17][7]=1.0; aaMut[17][8]=1.0; aaMut[17][9]=1.0; aaMut[17][10]=1.0; aaMut[17][11]=1.0; aaMut[17][12]=1.0; aaMut[17][13]=1.0; aaMut[17][14]=1.0; aaMut[17][15]=1.0; aaMut[17][16]=1.0; 
aaMut[18][0]=1.0; aaMut[18][1]=1.0; aaMut[18][2]=1.0; aaMut[18][3]=1.0; aaMut[18][4]=1.0; aaMut[18][5]=1.0; aaMut[18][6]=1.0; aaMut[18][7]=1.0; aaMut[18][8]=1.0; aaMut[18][9]=1.0; aaMut[18][10]=1.0; aaMut[18][11]=1.0; aaMut[18][12]=1.0; aaMut[18][13]=1.0; aaMut[18][14]=1.0; aaMut[18][15]=1.0; aaMut[18][16]=1.0; aaMut[18][17]=1.0; 
aaMut[19][0]=1.0; aaMut[19][1]=1.0; aaMut[19][2]=1.0; aaMut[19][3]=1.0; aaMut[19][4]=1.0; aaMut[19][5]=1.0; aaMut[19][6]=1.0; aaMut[19][7]=1.0; aaMut[19][8]=1.0; aaMut[19][9]=1.0; aaMut[19][10]=1.0; aaMut[19][11]=1.0; aaMut[19][12]=1.0; aaMut[19][13]=1.0; aaMut[19][14]=1.0; aaMut[19][15]=1.0; aaMut[19][16]=1.0; aaMut[19][17]=1.0; aaMut[19][18]=1.0; 


	// from triangular exchange matrix to square exchange  matrix
	for (i=0; i<20; i++) {
	  for (j=0; j<=i; j++) 
	    a[i][j] = aaMut[i][j];
	  
	  for (j=i+1; j<20; j++)
	    a[i][j] = aaMut[j][i];  
	}
	
	r = 0;
	for(i = 0; i < 19; i++)
	  {      
	    for(j = i + 1; j < 20; j++) 
	      initialRates[r++] = a[i][j];
	  }
	
      }

      /* F81 uses empirical aa frequencies from the alignment */
      for(i = 0; i < 20; i++)		
	f[i] = tr->frequencies[model * 20 + i];

      return (0);
      break;

    case DAYHOFF:
      {	
	 daa[ 1*20+ 0] =   27.00; daa[ 2*20+ 0] =   98.00; daa[ 2*20+ 1] =   32.00; daa[ 3*20+ 0] =  120.00;
	 daa[ 3*20+ 1] =    0.00; daa[ 3*20+ 2] =  905.00; daa[ 4*20+ 0] =   36.00; daa[ 4*20+ 1] =   23.00;
	 daa[ 4*20+ 2] =    0.00; daa[ 4*20+ 3] =    0.00; daa[ 5*20+ 0] =   89.00; daa[ 5*20+ 1] =  246.00;
	 daa[ 5*20+ 2] =  103.00; daa[ 5*20+ 3] =  134.00; daa[ 5*20+ 4] =    0.00; daa[ 6*20+ 0] =  198.00;
	 daa[ 6*20+ 1] =    1.00; daa[ 6*20+ 2] =  148.00; daa[ 6*20+ 3] = 1153.00; daa[ 6*20+ 4] =    0.00;
	 daa[ 6*20+ 5] =  716.00; daa[ 7*20+ 0] =  240.00; daa[ 7*20+ 1] =    9.00; daa[ 7*20+ 2] =  139.00;
	 daa[ 7*20+ 3] =  125.00; daa[ 7*20+ 4] =   11.00; daa[ 7*20+ 5] =   28.00; daa[ 7*20+ 6] =   81.00;
	 daa[ 8*20+ 0] =   23.00; daa[ 8*20+ 1] =  240.00; daa[ 8*20+ 2] =  535.00; daa[ 8*20+ 3] =   86.00;
	 daa[ 8*20+ 4] =   28.00; daa[ 8*20+ 5] =  606.00; daa[ 8*20+ 6] =   43.00; daa[ 8*20+ 7] =   10.00;
	 daa[ 9*20+ 0] =   65.00; daa[ 9*20+ 1] =   64.00; daa[ 9*20+ 2] =   77.00; daa[ 9*20+ 3] =   24.00;
	 daa[ 9*20+ 4] =   44.00; daa[ 9*20+ 5] =   18.00; daa[ 9*20+ 6] =   61.00; daa[ 9*20+ 7] =    0.00;
	 daa[ 9*20+ 8] =    7.00; daa[10*20+ 0] =   41.00; daa[10*20+ 1] =   15.00; daa[10*20+ 2] =   34.00;
	 daa[10*20+ 3] =    0.00; daa[10*20+ 4] =    0.00; daa[10*20+ 5] =   73.00; daa[10*20+ 6] =   11.00;
	 daa[10*20+ 7] =    7.00; daa[10*20+ 8] =   44.00; daa[10*20+ 9] =  257.00; daa[11*20+ 0] =   26.00;
	 daa[11*20+ 1] =  464.00; daa[11*20+ 2] =  318.00; daa[11*20+ 3] =   71.00; daa[11*20+ 4] =    0.00;
	 daa[11*20+ 5] =  153.00; daa[11*20+ 6] =   83.00; daa[11*20+ 7] =   27.00; daa[11*20+ 8] =   26.00;
	 daa[11*20+ 9] =   46.00; daa[11*20+10] =   18.00; daa[12*20+ 0] =   72.00; daa[12*20+ 1] =   90.00;
	 daa[12*20+ 2] =    1.00; daa[12*20+ 3] =    0.00; daa[12*20+ 4] =    0.00; daa[12*20+ 5] =  114.00;
	 daa[12*20+ 6] =   30.00; daa[12*20+ 7] =   17.00; daa[12*20+ 8] =    0.00; daa[12*20+ 9] =  336.00;
	 daa[12*20+10] =  527.00; daa[12*20+11] =  243.00; daa[13*20+ 0] =   18.00; daa[13*20+ 1] =   14.00;
	 daa[13*20+ 2] =   14.00; daa[13*20+ 3] =    0.00; daa[13*20+ 4] =    0.00; daa[13*20+ 5] =    0.00;
	 daa[13*20+ 6] =    0.00; daa[13*20+ 7] =   15.00; daa[13*20+ 8] =   48.00; daa[13*20+ 9] =  196.00;
	 daa[13*20+10] =  157.00; daa[13*20+11] =    0.00; daa[13*20+12] =   92.00; daa[14*20+ 0] =  250.00;
	 daa[14*20+ 1] =  103.00; daa[14*20+ 2] =   42.00; daa[14*20+ 3] =   13.00; daa[14*20+ 4] =   19.00;
	 daa[14*20+ 5] =  153.00; daa[14*20+ 6] =   51.00; daa[14*20+ 7] =   34.00; daa[14*20+ 8] =   94.00;
	 daa[14*20+ 9] =   12.00; daa[14*20+10] =   32.00; daa[14*20+11] =   33.00; daa[14*20+12] =   17.00;
	 daa[14*20+13] =   11.00; daa[15*20+ 0] =  409.00; daa[15*20+ 1] =  154.00; daa[15*20+ 2] =  495.00;
	 daa[15*20+ 3] =   95.00; daa[15*20+ 4] =  161.00; daa[15*20+ 5] =   56.00; daa[15*20+ 6] =   79.00;
	 daa[15*20+ 7] =  234.00; daa[15*20+ 8] =   35.00; daa[15*20+ 9] =   24.00; daa[15*20+10] =   17.00;
	 daa[15*20+11] =   96.00; daa[15*20+12] =   62.00; daa[15*20+13] =   46.00; daa[15*20+14] =  245.00;
	 daa[16*20+ 0] =  371.00; daa[16*20+ 1] =   26.00; daa[16*20+ 2] =  229.00; daa[16*20+ 3] =   66.00;
	 daa[16*20+ 4] =   16.00; daa[16*20+ 5] =   53.00; daa[16*20+ 6] =   34.00; daa[16*20+ 7] =   30.00;
	 daa[16*20+ 8] =   22.00; daa[16*20+ 9] =  192.00; daa[16*20+10] =   33.00; daa[16*20+11] =  136.00;
	 daa[16*20+12] =  104.00; daa[16*20+13] =   13.00; daa[16*20+14] =   78.00; daa[16*20+15] =  550.00;
	 daa[17*20+ 0] =    0.00; daa[17*20+ 1] =  201.00; daa[17*20+ 2] =   23.00; daa[17*20+ 3] =    0.00;
	 daa[17*20+ 4] =    0.00; daa[17*20+ 5] =    0.00; daa[17*20+ 6] =    0.00; daa[17*20+ 7] =    0.00;
	 daa[17*20+ 8] =   27.00; daa[17*20+ 9] =    0.00; daa[17*20+10] =   46.00; daa[17*20+11] =    0.00;
	 daa[17*20+12] =    0.00; daa[17*20+13] =   76.00; daa[17*20+14] =    0.00; daa[17*20+15] =   75.00;
	 daa[17*20+16] =    0.00; daa[18*20+ 0] =   24.00; daa[18*20+ 1] =    8.00; daa[18*20+ 2] =   95.00;
	 daa[18*20+ 3] =    0.00; daa[18*20+ 4] =   96.00; daa[18*20+ 5] =    0.00; daa[18*20+ 6] =   22.00;
	 daa[18*20+ 7] =    0.00; daa[18*20+ 8] =  127.00; daa[18*20+ 9] =   37.00; daa[18*20+10] =   28.00;
	 daa[18*20+11] =   13.00; daa[18*20+12] =    0.00; daa[18*20+13] =  698.00; daa[18*20+14] =    0.00;
	 daa[18*20+15] =   34.00; daa[18*20+16] =   42.00; daa[18*20+17] =   61.00; daa[19*20+ 0] =  208.00;
	 daa[19*20+ 1] =   24.00; daa[19*20+ 2] =   15.00; daa[19*20+ 3] =   18.00; daa[19*20+ 4] =   49.00;
	 daa[19*20+ 5] =   35.00; daa[19*20+ 6] =   37.00; daa[19*20+ 7] =   54.00; daa[19*20+ 8] =   44.00;
	 daa[19*20+ 9] =  889.00; daa[19*20+10] =  175.00; daa[19*20+11] =   10.00; daa[19*20+12] =  258.00;
	 daa[19*20+13] =   12.00; daa[19*20+14] =   48.00; daa[19*20+15] =   30.00; daa[19*20+16] =  157.00;
	 daa[19*20+17] =    0.00; daa[19*20+18] =   28.00;
	   
	 f[ 0] = 0.087127; f[ 1] = 0.040904; f[ 2] = 0.040432; f[ 3] = 0.046872;
	 f[ 4] = 0.033474; f[ 5] = 0.038255; f[ 6] = 0.049530; f[ 7] = 0.088612;
	 f[ 8] = 0.033618; f[ 9] = 0.036886; f[10] = 0.085357; f[11] = 0.080482;
	 f[12] = 0.014753; f[13] = 0.039772; f[14] = 0.050680; f[15] = 0.069577;
	 f[16] = 0.058542; f[17] = 0.010494; f[18] = 0.029916; f[19] = 0.064718;
      }
      break;
    case DCMUT:
      {	
	daa[ 1*20+ 0] =   26.78280; daa[ 2*20+ 0] =   98.44740; daa[ 2*20+ 1] =   32.70590; daa[ 3*20+ 0] =  119.98050; 
	daa[ 3*20+ 1] =    0.00000; daa[ 3*20+ 2] =  893.15150; daa[ 4*20+ 0] =   36.00160; daa[ 4*20+ 1] =   23.23740; 
	daa[ 4*20+ 2] =    0.00000; daa[ 4*20+ 3] =    0.00000; daa[ 5*20+ 0] =   88.77530; daa[ 5*20+ 1] =  243.99390; 
	daa[ 5*20+ 2] =  102.85090; daa[ 5*20+ 3] =  134.85510; daa[ 5*20+ 4] =    0.00000; daa[ 6*20+ 0] =  196.11670; 
	daa[ 6*20+ 1] =    0.00000; daa[ 6*20+ 2] =  149.34090; daa[ 6*20+ 3] = 1138.86590; daa[ 6*20+ 4] =    0.00000; 
	daa[ 6*20+ 5] =  708.60220; daa[ 7*20+ 0] =  238.61110; daa[ 7*20+ 1] =    8.77910; daa[ 7*20+ 2] =  138.53520; 
	daa[ 7*20+ 3] =  124.09810; daa[ 7*20+ 4] =   10.72780; daa[ 7*20+ 5] =   28.15810; daa[ 7*20+ 6] =   81.19070; 
	daa[ 8*20+ 0] =   22.81160; daa[ 8*20+ 1] =  238.31480; daa[ 8*20+ 2] =  529.00240; daa[ 8*20+ 3] =   86.82410; 
	daa[ 8*20+ 4] =   28.27290; daa[ 8*20+ 5] =  601.16130; daa[ 8*20+ 6] =   43.94690; daa[ 8*20+ 7] =   10.68020; 
	daa[ 9*20+ 0] =   65.34160; daa[ 9*20+ 1] =   63.26290; daa[ 9*20+ 2] =   76.80240; daa[ 9*20+ 3] =   23.92480; 
	daa[ 9*20+ 4] =   43.80740; daa[ 9*20+ 5] =   18.03930; daa[ 9*20+ 6] =   60.95260; daa[ 9*20+ 7] =    0.00000; 
	daa[ 9*20+ 8] =    7.69810; daa[10*20+ 0] =   40.64310; daa[10*20+ 1] =   15.49240; daa[10*20+ 2] =   34.11130; 
	daa[10*20+ 3] =    0.00000; daa[10*20+ 4] =    0.00000; daa[10*20+ 5] =   73.07720; daa[10*20+ 6] =   11.28800; 
	daa[10*20+ 7] =    7.15140; daa[10*20+ 8] =   44.35040; daa[10*20+ 9] =  255.66850; daa[11*20+ 0] =   25.86350; 
	daa[11*20+ 1] =  461.01240; daa[11*20+ 2] =  314.83710; daa[11*20+ 3] =   71.69130; daa[11*20+ 4] =    0.00000; 
	daa[11*20+ 5] =  151.90780; daa[11*20+ 6] =   83.00780; daa[11*20+ 7] =   26.76830; daa[11*20+ 8] =   27.04750; 
	daa[11*20+ 9] =   46.08570; daa[11*20+10] =   18.06290; daa[12*20+ 0] =   71.78400; daa[12*20+ 1] =   89.63210; 
	daa[12*20+ 2] =    0.00000; daa[12*20+ 3] =    0.00000; daa[12*20+ 4] =    0.00000; daa[12*20+ 5] =  112.74990; 
	daa[12*20+ 6] =   30.48030; daa[12*20+ 7] =   17.03720; daa[12*20+ 8] =    0.00000; daa[12*20+ 9] =  333.27320; 
	daa[12*20+10] =  523.01150; daa[12*20+11] =  241.17390; daa[13*20+ 0] =   18.36410; daa[13*20+ 1] =   13.69060; 
	daa[13*20+ 2] =   13.85030; daa[13*20+ 3] =    0.00000; daa[13*20+ 4] =    0.00000; daa[13*20+ 5] =    0.00000; 
	daa[13*20+ 6] =    0.00000; daa[13*20+ 7] =   15.34780; daa[13*20+ 8] =   47.59270; daa[13*20+ 9] =  195.19510; 
	daa[13*20+10] =  156.51600; daa[13*20+11] =    0.00000; daa[13*20+12] =   92.18600; daa[14*20+ 0] =  248.59200; 
	daa[14*20+ 1] =  102.83130; daa[14*20+ 2] =   41.92440; daa[14*20+ 3] =   13.39400; daa[14*20+ 4] =   18.75500; 
	daa[14*20+ 5] =  152.61880; daa[14*20+ 6] =   50.70030; daa[14*20+ 7] =   34.71530; daa[14*20+ 8] =   93.37090; 
	daa[14*20+ 9] =   11.91520; daa[14*20+10] =   31.62580; daa[14*20+11] =   33.54190; daa[14*20+12] =   17.02050; 
	daa[14*20+13] =   11.05060; daa[15*20+ 0] =  405.18700; daa[15*20+ 1] =  153.15900; daa[15*20+ 2] =  488.58920; 
	daa[15*20+ 3] =   95.60970; daa[15*20+ 4] =  159.83560; daa[15*20+ 5] =   56.18280; daa[15*20+ 6] =   79.39990; 
	daa[15*20+ 7] =  232.22430; daa[15*20+ 8] =   35.36430; daa[15*20+ 9] =   24.79550; daa[15*20+10] =   17.14320; 
	daa[15*20+11] =   95.45570; daa[15*20+12] =   61.99510; daa[15*20+13] =   45.99010; daa[15*20+14] =  242.72020; 
	daa[16*20+ 0] =  368.03650; daa[16*20+ 1] =   26.57450; daa[16*20+ 2] =  227.16970; daa[16*20+ 3] =   66.09300; 
	daa[16*20+ 4] =   16.23660; daa[16*20+ 5] =   52.56510; daa[16*20+ 6] =   34.01560; daa[16*20+ 7] =   30.66620; 
	daa[16*20+ 8] =   22.63330; daa[16*20+ 9] =  190.07390; daa[16*20+10] =   33.10900; daa[16*20+11] =  135.05990; 
	daa[16*20+12] =  103.15340; daa[16*20+13] =   13.66550; daa[16*20+14] =   78.28570; daa[16*20+15] =  543.66740; 
	daa[17*20+ 0] =    0.00000; daa[17*20+ 1] =  200.13750; daa[17*20+ 2] =   22.49680; daa[17*20+ 3] =    0.00000; 
	daa[17*20+ 4] =    0.00000; daa[17*20+ 5] =    0.00000; daa[17*20+ 6] =    0.00000; daa[17*20+ 7] =    0.00000; 
	daa[17*20+ 8] =   27.05640; daa[17*20+ 9] =    0.00000; daa[17*20+10] =   46.17760; daa[17*20+11] =    0.00000; 
	daa[17*20+12] =    0.00000; daa[17*20+13] =   76.23540; daa[17*20+14] =    0.00000; daa[17*20+15] =   74.08190; 
	daa[17*20+16] =    0.00000; daa[18*20+ 0] =   24.41390; daa[18*20+ 1] =    7.80120; daa[18*20+ 2] =   94.69400; 
	daa[18*20+ 3] =    0.00000; daa[18*20+ 4] =   95.31640; daa[18*20+ 5] =    0.00000; daa[18*20+ 6] =   21.47170; 
	daa[18*20+ 7] =    0.00000; daa[18*20+ 8] =  126.54000; daa[18*20+ 9] =   37.48340; daa[18*20+10] =   28.65720; 
	daa[18*20+11] =   13.21420; daa[18*20+12] =    0.00000; daa[18*20+13] =  695.26290; daa[18*20+14] =    0.00000; 
	daa[18*20+15] =   33.62890; daa[18*20+16] =   41.78390; daa[18*20+17] =   60.80700; daa[19*20+ 0] =  205.95640; 
	daa[19*20+ 1] =   24.03680; daa[19*20+ 2] =   15.80670; daa[19*20+ 3] =   17.83160; daa[19*20+ 4] =   48.46780; 
	daa[19*20+ 5] =   34.69830; daa[19*20+ 6] =   36.72500; daa[19*20+ 7] =   53.81650; daa[19*20+ 8] =   43.87150; 
	daa[19*20+ 9] =  881.00380; daa[19*20+10] =  174.51560; daa[19*20+11] =   10.38500; daa[19*20+12] =  256.59550; 
	daa[19*20+13] =   12.36060; daa[19*20+14] =   48.50260; daa[19*20+15] =   30.38360; daa[19*20+16] =  156.19970; 
	daa[19*20+17] =    0.00000; daa[19*20+18] =   27.93790;   
  
	f[ 0] = 0.087127; f[ 1] = 0.040904; f[ 2] = 0.040432; f[ 3] = 0.046872; 
	f[ 4] = 0.033474; f[ 5] = 0.038255; f[ 6] = 0.049530; f[ 7] = 0.088612;
	f[ 8] = 0.033619; f[ 9] = 0.036886; f[10] = 0.085357; f[11] = 0.080481;
	f[12] = 0.014753; f[13] = 0.039772; f[14] = 0.050680; f[15] = 0.069577;
	f[16] = 0.058542; f[17] = 0.010494; f[18] = 0.029916; f[19] = 0.064718;	
      }
      break;
    case JTT:
      {
	daa[ 1*20+ 0] =   58.00; daa[ 2*20+ 0] =   54.00; daa[ 2*20+ 1] =   45.00; daa[ 3*20+ 0] =   81.00;
	daa[ 3*20+ 1] =   16.00; daa[ 3*20+ 2] =  528.00; daa[ 4*20+ 0] =   56.00; daa[ 4*20+ 1] =  113.00;
	daa[ 4*20+ 2] =   34.00; daa[ 4*20+ 3] =   10.00; daa[ 5*20+ 0] =   57.00; daa[ 5*20+ 1] =  310.00;
	daa[ 5*20+ 2] =   86.00; daa[ 5*20+ 3] =   49.00; daa[ 5*20+ 4] =    9.00; daa[ 6*20+ 0] =  105.00;
	daa[ 6*20+ 1] =   29.00; daa[ 6*20+ 2] =   58.00; daa[ 6*20+ 3] =  767.00; daa[ 6*20+ 4] =    5.00;
	daa[ 6*20+ 5] =  323.00; daa[ 7*20+ 0] =  179.00; daa[ 7*20+ 1] =  137.00; daa[ 7*20+ 2] =   81.00;
	daa[ 7*20+ 3] =  130.00; daa[ 7*20+ 4] =   59.00; daa[ 7*20+ 5] =   26.00; daa[ 7*20+ 6] =  119.00;
	daa[ 8*20+ 0] =   27.00; daa[ 8*20+ 1] =  328.00; daa[ 8*20+ 2] =  391.00; daa[ 8*20+ 3] =  112.00;
	daa[ 8*20+ 4] =   69.00; daa[ 8*20+ 5] =  597.00; daa[ 8*20+ 6] =   26.00; daa[ 8*20+ 7] =   23.00;
	daa[ 9*20+ 0] =   36.00; daa[ 9*20+ 1] =   22.00; daa[ 9*20+ 2] =   47.00; daa[ 9*20+ 3] =   11.00;
	daa[ 9*20+ 4] =   17.00; daa[ 9*20+ 5] =    9.00; daa[ 9*20+ 6] =   12.00; daa[ 9*20+ 7] =    6.00;
	daa[ 9*20+ 8] =   16.00; daa[10*20+ 0] =   30.00; daa[10*20+ 1] =   38.00; daa[10*20+ 2] =   12.00;
	daa[10*20+ 3] =    7.00; daa[10*20+ 4] =   23.00; daa[10*20+ 5] =   72.00; daa[10*20+ 6] =    9.00;
	daa[10*20+ 7] =    6.00; daa[10*20+ 8] =   56.00; daa[10*20+ 9] =  229.00; daa[11*20+ 0] =   35.00;
	daa[11*20+ 1] =  646.00; daa[11*20+ 2] =  263.00; daa[11*20+ 3] =   26.00; daa[11*20+ 4] =    7.00;
	daa[11*20+ 5] =  292.00; daa[11*20+ 6] =  181.00; daa[11*20+ 7] =   27.00; daa[11*20+ 8] =   45.00;
	daa[11*20+ 9] =   21.00; daa[11*20+10] =   14.00; daa[12*20+ 0] =   54.00; daa[12*20+ 1] =   44.00;
	daa[12*20+ 2] =   30.00; daa[12*20+ 3] =   15.00; daa[12*20+ 4] =   31.00; daa[12*20+ 5] =   43.00;
	daa[12*20+ 6] =   18.00; daa[12*20+ 7] =   14.00; daa[12*20+ 8] =   33.00; daa[12*20+ 9] =  479.00;
	daa[12*20+10] =  388.00; daa[12*20+11] =   65.00; daa[13*20+ 0] =   15.00; daa[13*20+ 1] =    5.00;
	daa[13*20+ 2] =   10.00; daa[13*20+ 3] =    4.00; daa[13*20+ 4] =   78.00; daa[13*20+ 5] =    4.00;
	daa[13*20+ 6] =    5.00; daa[13*20+ 7] =    5.00; daa[13*20+ 8] =   40.00; daa[13*20+ 9] =   89.00;
	daa[13*20+10] =  248.00; daa[13*20+11] =    4.00; daa[13*20+12] =   43.00; daa[14*20+ 0] =  194.00;
	daa[14*20+ 1] =   74.00; daa[14*20+ 2] =   15.00; daa[14*20+ 3] =   15.00; daa[14*20+ 4] =   14.00;
	daa[14*20+ 5] =  164.00; daa[14*20+ 6] =   18.00; daa[14*20+ 7] =   24.00; daa[14*20+ 8] =  115.00;
	daa[14*20+ 9] =   10.00; daa[14*20+10] =  102.00; daa[14*20+11] =   21.00; daa[14*20+12] =   16.00;
	daa[14*20+13] =   17.00; daa[15*20+ 0] =  378.00; daa[15*20+ 1] =  101.00; daa[15*20+ 2] =  503.00;
	daa[15*20+ 3] =   59.00; daa[15*20+ 4] =  223.00; daa[15*20+ 5] =   53.00; daa[15*20+ 6] =   30.00;
	daa[15*20+ 7] =  201.00; daa[15*20+ 8] =   73.00; daa[15*20+ 9] =   40.00; daa[15*20+10] =   59.00;
	daa[15*20+11] =   47.00; daa[15*20+12] =   29.00; daa[15*20+13] =   92.00; daa[15*20+14] =  285.00;
	daa[16*20+ 0] =  475.00; daa[16*20+ 1] =   64.00; daa[16*20+ 2] =  232.00; daa[16*20+ 3] =   38.00;
	daa[16*20+ 4] =   42.00; daa[16*20+ 5] =   51.00; daa[16*20+ 6] =   32.00; daa[16*20+ 7] =   33.00;
	daa[16*20+ 8] =   46.00; daa[16*20+ 9] =  245.00; daa[16*20+10] =   25.00; daa[16*20+11] =  103.00;
	daa[16*20+12] =  226.00; daa[16*20+13] =   12.00; daa[16*20+14] =  118.00; daa[16*20+15] =  477.00;
	daa[17*20+ 0] =    9.00; daa[17*20+ 1] =  126.00; daa[17*20+ 2] =    8.00; daa[17*20+ 3] =    4.00;
	daa[17*20+ 4] =  115.00; daa[17*20+ 5] =   18.00; daa[17*20+ 6] =   10.00; daa[17*20+ 7] =   55.00;
	daa[17*20+ 8] =    8.00; daa[17*20+ 9] =    9.00; daa[17*20+10] =   52.00; daa[17*20+11] =   10.00;
	daa[17*20+12] =   24.00; daa[17*20+13] =   53.00; daa[17*20+14] =    6.00; daa[17*20+15] =   35.00;
	daa[17*20+16] =   12.00; daa[18*20+ 0] =   11.00; daa[18*20+ 1] =   20.00; daa[18*20+ 2] =   70.00;
	daa[18*20+ 3] =   46.00; daa[18*20+ 4] =  209.00; daa[18*20+ 5] =   24.00; daa[18*20+ 6] =    7.00;
	daa[18*20+ 7] =    8.00; daa[18*20+ 8] =  573.00; daa[18*20+ 9] =   32.00; daa[18*20+10] =   24.00;
	daa[18*20+11] =    8.00; daa[18*20+12] =   18.00; daa[18*20+13] =  536.00; daa[18*20+14] =   10.00;
	daa[18*20+15] =   63.00; daa[18*20+16] =   21.00; daa[18*20+17] =   71.00; daa[19*20+ 0] =  298.00;
	daa[19*20+ 1] =   17.00; daa[19*20+ 2] =   16.00; daa[19*20+ 3] =   31.00; daa[19*20+ 4] =   62.00;
	daa[19*20+ 5] =   20.00; daa[19*20+ 6] =   45.00; daa[19*20+ 7] =   47.00; daa[19*20+ 8] =   11.00;
	daa[19*20+ 9] =  961.00; daa[19*20+10] =  180.00; daa[19*20+11] =   14.00; daa[19*20+12] =  323.00;
	daa[19*20+13] =   62.00; daa[19*20+14] =   23.00; daa[19*20+15] =   38.00; daa[19*20+16] =  112.00;
	daa[19*20+17] =   25.00; daa[19*20+18] =   16.00;

	f[ 0] = 0.076748; f[ 1] = 0.051691; f[ 2] = 0.042645; f[ 3] = 0.051544;
	f[ 4] = 0.019803; f[ 5] = 0.040752; f[ 6] = 0.061830; f[ 7] = 0.073152;
	f[ 8] = 0.022944; f[ 9] = 0.053761; f[10] = 0.091904; f[11] = 0.058676;
	f[12] = 0.023826; f[13] = 0.040126; f[14] = 0.050901; f[15] = 0.068765;
	f[16] = 0.058565; f[17] = 0.014261; f[18] = 0.032102; f[19] = 0.066005;
      }
      break;
    case  MTREV:
      {
	daa[ 1*20+ 0] =   23.18; daa[ 2*20+ 0] =   26.95; daa[ 2*20+ 1] =   13.24; daa[ 3*20+ 0] =   17.67;
	daa[ 3*20+ 1] =    1.90; daa[ 3*20+ 2] =  794.38; daa[ 4*20+ 0] =   59.93; daa[ 4*20+ 1] =  103.33;
	daa[ 4*20+ 2] =   58.94; daa[ 4*20+ 3] =    1.90; daa[ 5*20+ 0] =    1.90; daa[ 5*20+ 1] =  220.99;
	daa[ 5*20+ 2] =  173.56; daa[ 5*20+ 3] =   55.28; daa[ 5*20+ 4] =   75.24; daa[ 6*20+ 0] =    9.77;
	daa[ 6*20+ 1] =    1.90; daa[ 6*20+ 2] =   63.05; daa[ 6*20+ 3] =  583.55; daa[ 6*20+ 4] =    1.90;
	daa[ 6*20+ 5] =  313.56; daa[ 7*20+ 0] =  120.71; daa[ 7*20+ 1] =   23.03; daa[ 7*20+ 2] =   53.30;
	daa[ 7*20+ 3] =   56.77; daa[ 7*20+ 4] =   30.71; daa[ 7*20+ 5] =    6.75; daa[ 7*20+ 6] =   28.28;
	daa[ 8*20+ 0] =   13.90; daa[ 8*20+ 1] =  165.23; daa[ 8*20+ 2] =  496.13; daa[ 8*20+ 3] =  113.99;
	daa[ 8*20+ 4] =  141.49; daa[ 8*20+ 5] =  582.40; daa[ 8*20+ 6] =   49.12; daa[ 8*20+ 7] =    1.90;
	daa[ 9*20+ 0] =   96.49; daa[ 9*20+ 1] =    1.90; daa[ 9*20+ 2] =   27.10; daa[ 9*20+ 3] =    4.34;
	daa[ 9*20+ 4] =   62.73; daa[ 9*20+ 5] =    8.34; daa[ 9*20+ 6] =    3.31; daa[ 9*20+ 7] =    5.98;
	daa[ 9*20+ 8] =   12.26; daa[10*20+ 0] =   25.46; daa[10*20+ 1] =   15.58; daa[10*20+ 2] =   15.16;
	daa[10*20+ 3] =    1.90; daa[10*20+ 4] =   25.65; daa[10*20+ 5] =   39.70; daa[10*20+ 6] =    1.90;
	daa[10*20+ 7] =    2.41; daa[10*20+ 8] =   11.49; daa[10*20+ 9] =  329.09; daa[11*20+ 0] =    8.36;
	daa[11*20+ 1] =  141.40; daa[11*20+ 2] =  608.70; daa[11*20+ 3] =    2.31; daa[11*20+ 4] =    1.90;
	daa[11*20+ 5] =  465.58; daa[11*20+ 6] =  313.86; daa[11*20+ 7] =   22.73; daa[11*20+ 8] =  127.67;
	daa[11*20+ 9] =   19.57; daa[11*20+10] =   14.88; daa[12*20+ 0] =  141.88; daa[12*20+ 1] =    1.90;
	daa[12*20+ 2] =   65.41; daa[12*20+ 3] =    1.90; daa[12*20+ 4] =    6.18; daa[12*20+ 5] =   47.37;
	daa[12*20+ 6] =    1.90; daa[12*20+ 7] =    1.90; daa[12*20+ 8] =   11.97; daa[12*20+ 9] =  517.98;
	daa[12*20+10] =  537.53; daa[12*20+11] =   91.37; daa[13*20+ 0] =    6.37; daa[13*20+ 1] =    4.69;
	daa[13*20+ 2] =   15.20; daa[13*20+ 3] =    4.98; daa[13*20+ 4] =   70.80; daa[13*20+ 5] =   19.11;
	daa[13*20+ 6] =    2.67; daa[13*20+ 7] =    1.90; daa[13*20+ 8] =   48.16; daa[13*20+ 9] =   84.67;
	daa[13*20+10] =  216.06; daa[13*20+11] =    6.44; daa[13*20+12] =   90.82; daa[14*20+ 0] =   54.31;
	daa[14*20+ 1] =   23.64; daa[14*20+ 2] =   73.31; daa[14*20+ 3] =   13.43; daa[14*20+ 4] =   31.26;
	daa[14*20+ 5] =  137.29; daa[14*20+ 6] =   12.83; daa[14*20+ 7] =    1.90; daa[14*20+ 8] =   60.97;
	daa[14*20+ 9] =   20.63; daa[14*20+10] =   40.10; daa[14*20+11] =   50.10; daa[14*20+12] =   18.84;
	daa[14*20+13] =   17.31; daa[15*20+ 0] =  387.86; daa[15*20+ 1] =    6.04; daa[15*20+ 2] =  494.39;
	daa[15*20+ 3] =   69.02; daa[15*20+ 4] =  277.05; daa[15*20+ 5] =   54.11; daa[15*20+ 6] =   54.71;
	daa[15*20+ 7] =  125.93; daa[15*20+ 8] =   77.46; daa[15*20+ 9] =   47.70; daa[15*20+10] =   73.61;
	daa[15*20+11] =  105.79; daa[15*20+12] =  111.16; daa[15*20+13] =   64.29; daa[15*20+14] =  169.90;
	daa[16*20+ 0] =  480.72; daa[16*20+ 1] =    2.08; daa[16*20+ 2] =  238.46; daa[16*20+ 3] =   28.01;
	daa[16*20+ 4] =  179.97; daa[16*20+ 5] =   94.93; daa[16*20+ 6] =   14.82; daa[16*20+ 7] =   11.17;
	daa[16*20+ 8] =   44.78; daa[16*20+ 9] =  368.43; daa[16*20+10] =  126.40; daa[16*20+11] =  136.33;
	daa[16*20+12] =  528.17; daa[16*20+13] =   33.85; daa[16*20+14] =  128.22; daa[16*20+15] =  597.21;
	daa[17*20+ 0] =    1.90; daa[17*20+ 1] =   21.95; daa[17*20+ 2] =   10.68; daa[17*20+ 3] =   19.86;
	daa[17*20+ 4] =   33.60; daa[17*20+ 5] =    1.90; daa[17*20+ 6] =    1.90; daa[17*20+ 7] =   10.92;
	daa[17*20+ 8] =    7.08; daa[17*20+ 9] =    1.90; daa[17*20+10] =   32.44; daa[17*20+11] =   24.00;
	daa[17*20+12] =   21.71; daa[17*20+13] =    7.84; daa[17*20+14] =    4.21; daa[17*20+15] =   38.58;
	daa[17*20+16] =    9.99; daa[18*20+ 0] =    6.48; daa[18*20+ 1] =    1.90; daa[18*20+ 2] =  191.36;
	daa[18*20+ 3] =   21.21; daa[18*20+ 4] =  254.77; daa[18*20+ 5] =   38.82; daa[18*20+ 6] =   13.12;
	daa[18*20+ 7] =    3.21; daa[18*20+ 8] =  670.14; daa[18*20+ 9] =   25.01; daa[18*20+10] =   44.15;
	daa[18*20+11] =   51.17; daa[18*20+12] =   39.96; daa[18*20+13] =  465.58; daa[18*20+14] =   16.21;
	daa[18*20+15] =   64.92; daa[18*20+16] =   38.73; daa[18*20+17] =   26.25; daa[19*20+ 0] =  195.06;
	daa[19*20+ 1] =    7.64; daa[19*20+ 2] =    1.90; daa[19*20+ 3] =    1.90; daa[19*20+ 4] =    1.90;
	daa[19*20+ 5] =   19.00; daa[19*20+ 6] =   21.14; daa[19*20+ 7] =    2.53; daa[19*20+ 8] =    1.90;
	daa[19*20+ 9] = 1222.94; daa[19*20+10] =   91.67; daa[19*20+11] =    1.90; daa[19*20+12] =  387.54;
	daa[19*20+13] =    6.35; daa[19*20+14] =    8.23; daa[19*20+15] =    1.90; daa[19*20+16] =  204.54;
	daa[19*20+17] =    5.37; daa[19*20+18] =    1.90;
	
  
	f[ 0] = 0.072000; f[ 1] = 0.019000; f[ 2] = 0.039000; f[ 3] = 0.019000;
	f[ 4] = 0.006000; f[ 5] = 0.025000; f[ 6] = 0.024000; f[ 7] = 0.056000;
	f[ 8] = 0.028000; f[ 9] = 0.088000; f[10] = 0.169000; f[11] = 0.023000;
	f[12] = 0.054000; f[13] = 0.061000; f[14] = 0.054000; f[15] = 0.072000;
	f[16] = 0.086000; f[17] = 0.029000; f[18] = 0.033000; f[19] = 0.043000;
      }
      break;
    case WAG:
      {
	daa[ 1*20+ 0] =  55.15710; daa[ 2*20+ 0] =  50.98480; daa[ 2*20+ 1] =  63.53460; 
	daa[ 3*20+ 0] =  73.89980; daa[ 3*20+ 1] =  14.73040; daa[ 3*20+ 2] = 542.94200; 
	daa[ 4*20+ 0] = 102.70400; daa[ 4*20+ 1] =  52.81910; daa[ 4*20+ 2] =  26.52560; 
	daa[ 4*20+ 3] =   3.02949; daa[ 5*20+ 0] =  90.85980; daa[ 5*20+ 1] = 303.55000; 
	daa[ 5*20+ 2] = 154.36400; daa[ 5*20+ 3] =  61.67830; daa[ 5*20+ 4] =   9.88179; 
	daa[ 6*20+ 0] = 158.28500; daa[ 6*20+ 1] =  43.91570; daa[ 6*20+ 2] =  94.71980; 
	daa[ 6*20+ 3] = 617.41600; daa[ 6*20+ 4] =   2.13520; daa[ 6*20+ 5] = 546.94700; 
	daa[ 7*20+ 0] = 141.67200; daa[ 7*20+ 1] =  58.46650; daa[ 7*20+ 2] = 112.55600; 
	daa[ 7*20+ 3] =  86.55840; daa[ 7*20+ 4] =  30.66740; daa[ 7*20+ 5] =  33.00520; 
	daa[ 7*20+ 6] =  56.77170; daa[ 8*20+ 0] =  31.69540; daa[ 8*20+ 1] = 213.71500; 
	daa[ 8*20+ 2] = 395.62900; daa[ 8*20+ 3] =  93.06760; daa[ 8*20+ 4] =  24.89720; 
	daa[ 8*20+ 5] = 429.41100; daa[ 8*20+ 6] =  57.00250; daa[ 8*20+ 7] =  24.94100; 
	daa[ 9*20+ 0] =  19.33350; daa[ 9*20+ 1] =  18.69790; daa[ 9*20+ 2] =  55.42360; 
	daa[ 9*20+ 3] =   3.94370; daa[ 9*20+ 4] =  17.01350; daa[ 9*20+ 5] =  11.39170; 
	daa[ 9*20+ 6] =  12.73950; daa[ 9*20+ 7] =   3.04501; daa[ 9*20+ 8] =  13.81900; 
	daa[10*20+ 0] =  39.79150; daa[10*20+ 1] =  49.76710; daa[10*20+ 2] =  13.15280; 
	daa[10*20+ 3] =   8.48047; daa[10*20+ 4] =  38.42870; daa[10*20+ 5] =  86.94890; 
	daa[10*20+ 6] =  15.42630; daa[10*20+ 7] =   6.13037; daa[10*20+ 8] =  49.94620; 
	daa[10*20+ 9] = 317.09700; daa[11*20+ 0] =  90.62650; daa[11*20+ 1] = 535.14200; 
	daa[11*20+ 2] = 301.20100; daa[11*20+ 3] =  47.98550; daa[11*20+ 4] =   7.40339; 
	daa[11*20+ 5] = 389.49000; daa[11*20+ 6] = 258.44300; daa[11*20+ 7] =  37.35580; 
	daa[11*20+ 8] =  89.04320; daa[11*20+ 9] =  32.38320; daa[11*20+10] =  25.75550; 
	daa[12*20+ 0] =  89.34960; daa[12*20+ 1] =  68.31620; daa[12*20+ 2] =  19.82210; 
	daa[12*20+ 3] =  10.37540; daa[12*20+ 4] =  39.04820; daa[12*20+ 5] = 154.52600; 
	daa[12*20+ 6] =  31.51240; daa[12*20+ 7] =  17.41000; daa[12*20+ 8] =  40.41410; 
	daa[12*20+ 9] = 425.74600; daa[12*20+10] = 485.40200; daa[12*20+11] =  93.42760; 
	daa[13*20+ 0] =  21.04940; daa[13*20+ 1] =  10.27110; daa[13*20+ 2] =   9.61621; 
	daa[13*20+ 3] =   4.67304; daa[13*20+ 4] =  39.80200; daa[13*20+ 5] =   9.99208; 
	daa[13*20+ 6] =   8.11339; daa[13*20+ 7] =   4.99310; daa[13*20+ 8] =  67.93710; 
	daa[13*20+ 9] = 105.94700; daa[13*20+10] = 211.51700; daa[13*20+11] =   8.88360; 
	daa[13*20+12] = 119.06300; daa[14*20+ 0] = 143.85500; daa[14*20+ 1] =  67.94890; 
	daa[14*20+ 2] =  19.50810; daa[14*20+ 3] =  42.39840; daa[14*20+ 4] =  10.94040; 
	daa[14*20+ 5] =  93.33720; daa[14*20+ 6] =  68.23550; daa[14*20+ 7] =  24.35700; 
	daa[14*20+ 8] =  69.61980; daa[14*20+ 9] =   9.99288; daa[14*20+10] =  41.58440; 
	daa[14*20+11] =  55.68960; daa[14*20+12] =  17.13290; daa[14*20+13] =  16.14440; 
	daa[15*20+ 0] = 337.07900; daa[15*20+ 1] = 122.41900; daa[15*20+ 2] = 397.42300; 
	daa[15*20+ 3] = 107.17600; daa[15*20+ 4] = 140.76600; daa[15*20+ 5] = 102.88700; 
	daa[15*20+ 6] =  70.49390; daa[15*20+ 7] = 134.18200; daa[15*20+ 8] =  74.01690; 
	daa[15*20+ 9] =  31.94400; daa[15*20+10] =  34.47390; daa[15*20+11] =  96.71300; 
	daa[15*20+12] =  49.39050; daa[15*20+13] =  54.59310; daa[15*20+14] = 161.32800; 
	daa[16*20+ 0] = 212.11100; daa[16*20+ 1] =  55.44130; daa[16*20+ 2] = 203.00600; 
	daa[16*20+ 3] =  37.48660; daa[16*20+ 4] =  51.29840; daa[16*20+ 5] =  85.79280; 
	daa[16*20+ 6] =  82.27650; daa[16*20+ 7] =  22.58330; daa[16*20+ 8] =  47.33070; 
	daa[16*20+ 9] = 145.81600; daa[16*20+10] =  32.66220; daa[16*20+11] = 138.69800; 
	daa[16*20+12] = 151.61200; daa[16*20+13] =  17.19030; daa[16*20+14] =  79.53840; 
	daa[16*20+15] = 437.80200; daa[17*20+ 0] =  11.31330; daa[17*20+ 1] = 116.39200; 
	daa[17*20+ 2] =   7.19167; daa[17*20+ 3] =  12.97670; daa[17*20+ 4] =  71.70700; 
	daa[17*20+ 5] =  21.57370; daa[17*20+ 6] =  15.65570; daa[17*20+ 7] =  33.69830; 
	daa[17*20+ 8] =  26.25690; daa[17*20+ 9] =  21.24830; daa[17*20+10] =  66.53090; 
	daa[17*20+11] =  13.75050; daa[17*20+12] =  51.57060; daa[17*20+13] = 152.96400; 
	daa[17*20+14] =  13.94050; daa[17*20+15] =  52.37420; daa[17*20+16] =  11.08640; 
	daa[18*20+ 0] =  24.07350; daa[18*20+ 1] =  38.15330; daa[18*20+ 2] = 108.60000; 
	daa[18*20+ 3] =  32.57110; daa[18*20+ 4] =  54.38330; daa[18*20+ 5] =  22.77100; 
	daa[18*20+ 6] =  19.63030; daa[18*20+ 7] =  10.36040; daa[18*20+ 8] = 387.34400; 
	daa[18*20+ 9] =  42.01700; daa[18*20+10] =  39.86180; daa[18*20+11] =  13.32640; 
	daa[18*20+12] =  42.84370; daa[18*20+13] = 645.42800; daa[18*20+14] =  21.60460; 
	daa[18*20+15] =  78.69930; daa[18*20+16] =  29.11480; daa[18*20+17] = 248.53900; 
	daa[19*20+ 0] = 200.60100; daa[19*20+ 1] =  25.18490; daa[19*20+ 2] =  19.62460; 
	daa[19*20+ 3] =  15.23350; daa[19*20+ 4] = 100.21400; daa[19*20+ 5] =  30.12810; 
	daa[19*20+ 6] =  58.87310; daa[19*20+ 7] =  18.72470; daa[19*20+ 8] =  11.83580; 
	daa[19*20+ 9] = 782.13000; daa[19*20+10] = 180.03400; daa[19*20+11] =  30.54340; 
	daa[19*20+12] = 205.84500; daa[19*20+13] =  64.98920; daa[19*20+14] =  31.48870; 
	daa[19*20+15] =  23.27390; daa[19*20+16] = 138.82300; daa[19*20+17] =  36.53690; 
	daa[19*20+18] =  31.47300; 
  
	f[0] = 0.0866279; f[1] =  0.043972; f[2] =  0.0390894; f[3] =  0.0570451;
	f[4] =  0.0193078; f[5] =  0.0367281; f[6] =  0.0580589; f[7] =  0.0832518;
	f[8] =  0.0244313; f[9] =  0.048466; f[10] =  0.086209; f[11] = 0.0620286;
	f[12] = 0.0195027; f[13] =  0.0384319; f[14] =  0.0457631; f[15] = 0.0695179;
	f[16] =  0.0610127; f[17] =  0.0143859; f[18] =  0.0352742; f[19] =  0.0708956;
      }
      break;
    case RTREV:
      {
	daa[1*20+0]= 34;         daa[2*20+0]= 51;         daa[2*20+1]= 35;         daa[3*20+0]= 10;         
	daa[3*20+1]= 30;         daa[3*20+2]= 384;        daa[4*20+0]= 439;        daa[4*20+1]= 92;         
	daa[4*20+2]= 128;        daa[4*20+3]= 1;          daa[5*20+0]= 32;         daa[5*20+1]= 221;        
	daa[5*20+2]= 236;        daa[5*20+3]= 78;         daa[5*20+4]= 70;         daa[6*20+0]= 81;         
	daa[6*20+1]= 10;         daa[6*20+2]= 79;         daa[6*20+3]= 542;        daa[6*20+4]= 1;          
	daa[6*20+5]= 372;        daa[7*20+0]= 135;        daa[7*20+1]= 41;         daa[7*20+2]= 94;         
	daa[7*20+3]= 61;         daa[7*20+4]= 48;         daa[7*20+5]= 18;         daa[7*20+6]= 70;         
	daa[8*20+0]= 30;         daa[8*20+1]= 90;         daa[8*20+2]= 320;        daa[8*20+3]= 91;         
	daa[8*20+4]= 124;        daa[8*20+5]= 387;        daa[8*20+6]= 34;         daa[8*20+7]= 68;         
	daa[9*20+0]= 1;          daa[9*20+1]= 24;         daa[9*20+2]= 35;         daa[9*20+3]= 1;          
	daa[9*20+4]= 104;        daa[9*20+5]= 33;         daa[9*20+6]= 1;          daa[9*20+7]= 1;          
	daa[9*20+8]= 34;         daa[10*20+0]= 45;        daa[10*20+1]= 18;        daa[10*20+2]= 15;        
	daa[10*20+3]= 5;         daa[10*20+4]= 110;       daa[10*20+5]= 54;        daa[10*20+6]= 21;        
	daa[10*20+7]= 3;         daa[10*20+8]= 51;        daa[10*20+9]= 385;       daa[11*20+0]= 38;        
	daa[11*20+1]= 593;       daa[11*20+2]= 123;       daa[11*20+3]= 20;        daa[11*20+4]= 16;        
	daa[11*20+5]= 309;       daa[11*20+6]= 141;       daa[11*20+7]= 30;        daa[11*20+8]= 76;        
	daa[11*20+9]= 34;        daa[11*20+10]= 23;       daa[12*20+0]= 235;       daa[12*20+1]= 57;        
	daa[12*20+2]= 1;         daa[12*20+3]= 1;         daa[12*20+4]= 156;       daa[12*20+5]= 158;       
	daa[12*20+6]= 1;         daa[12*20+7]= 37;        daa[12*20+8]= 116;       daa[12*20+9]= 375;       
	daa[12*20+10]= 581;      daa[12*20+11]= 134;      daa[13*20+0]= 1;         daa[13*20+1]= 7;         
	daa[13*20+2]= 49;        daa[13*20+3]= 1;         daa[13*20+4]= 70;        daa[13*20+5]= 1;         
	daa[13*20+6]= 1;         daa[13*20+7]= 7;         daa[13*20+8]= 141;       daa[13*20+9]= 64;        
	daa[13*20+10]= 179;      daa[13*20+11]= 14;       daa[13*20+12]= 247;      daa[14*20+0]= 97;        
	daa[14*20+1]= 24;        daa[14*20+2]= 33;        daa[14*20+3]= 55;        daa[14*20+4]= 1;         
	daa[14*20+5]= 68;        daa[14*20+6]= 52;        daa[14*20+7]= 17;        daa[14*20+8]= 44;        
	daa[14*20+9]= 10;        daa[14*20+10]= 22;       daa[14*20+11]= 43;       daa[14*20+12]= 1;        
	daa[14*20+13]= 11;       daa[15*20+0]= 460;       daa[15*20+1]= 102;       daa[15*20+2]= 294;       
	daa[15*20+3]= 136;       daa[15*20+4]= 75;        daa[15*20+5]= 225;       daa[15*20+6]= 95;        
	daa[15*20+7]= 152;       daa[15*20+8]= 183;       daa[15*20+9]= 4;         daa[15*20+10]= 24;       
	daa[15*20+11]= 77;       daa[15*20+12]= 1;        daa[15*20+13]= 20;       daa[15*20+14]= 134;      
	daa[16*20+0]= 258;       daa[16*20+1]= 64;        daa[16*20+2]= 148;       daa[16*20+3]= 55;        
	daa[16*20+4]= 117;       daa[16*20+5]= 146;       daa[16*20+6]= 82;        daa[16*20+7]= 7;         
	daa[16*20+8]= 49;        daa[16*20+9]= 72;        daa[16*20+10]= 25;       daa[16*20+11]= 110;      
	daa[16*20+12]= 131;      daa[16*20+13]= 69;       daa[16*20+14]= 62;       daa[16*20+15]= 671;      
	daa[17*20+0]= 5;         daa[17*20+1]= 13;        daa[17*20+2]= 16;        daa[17*20+3]= 1;         
	daa[17*20+4]= 55;        daa[17*20+5]= 10;        daa[17*20+6]= 17;        daa[17*20+7]= 23;        
	daa[17*20+8]= 48;        daa[17*20+9]= 39;        daa[17*20+10]= 47;       daa[17*20+11]= 6;        
	daa[17*20+12]= 111;      daa[17*20+13]= 182;      daa[17*20+14]= 9;        daa[17*20+15]= 14;       
	daa[17*20+16]= 1;        daa[18*20+0]= 55;        daa[18*20+1]= 47;        daa[18*20+2]= 28;        
	daa[18*20+3]= 1;         daa[18*20+4]= 131;       daa[18*20+5]= 45;        daa[18*20+6]= 1;         
	daa[18*20+7]= 21;        daa[18*20+8]= 307;       daa[18*20+9]= 26;        daa[18*20+10]= 64;       
	daa[18*20+11]= 1;        daa[18*20+12]= 74;       daa[18*20+13]= 1017;     daa[18*20+14]= 14;       
	daa[18*20+15]= 31;       daa[18*20+16]= 34;       daa[18*20+17]= 176;      daa[19*20+0]= 197;       
	daa[19*20+1]= 29;        daa[19*20+2]= 21;        daa[19*20+3]= 6;         daa[19*20+4]= 295;       
	daa[19*20+5]= 36;        daa[19*20+6]= 35;        daa[19*20+7]= 3;         daa[19*20+8]= 1;         
	daa[19*20+9]= 1048;      daa[19*20+10]= 112;      daa[19*20+11]= 19;       daa[19*20+12]= 236;      
	daa[19*20+13]= 92;       daa[19*20+14]= 25;       daa[19*20+15]= 39;       daa[19*20+16]= 196;      
	daa[19*20+17]= 26;       daa[19*20+18]= 59;       
			
	f[0]= 0.0646;           f[1]= 0.0453;           f[2]= 0.0376;           f[3]= 0.0422;           
	f[4]= 0.0114;           f[5]= 0.0606;           f[6]= 0.0607;           f[7]= 0.0639;           
	f[8]= 0.0273;           f[9]= 0.0679;           f[10]= 0.1018;          f[11]= 0.0751;          
	f[12]= 0.015;           f[13]= 0.0287;          f[14]= 0.0681;          f[15]= 0.0488;          
	f[16]= 0.0622;          f[17]= 0.0251;          f[18]= 0.0318;          f[19]= 0.0619;
      }
      break;
    case CPREV:
      {
	daa[1*20+0]= 105;        daa[2*20+0]= 227;        daa[2*20+1]= 357;        daa[3*20+0]= 175;        
	daa[3*20+1]= 43;         daa[3*20+2]= 4435;       daa[4*20+0]= 669;        daa[4*20+1]= 823;        
	daa[4*20+2]= 538;        daa[4*20+3]= 10;         daa[5*20+0]= 157;        daa[5*20+1]= 1745;       
	daa[5*20+2]= 768;        daa[5*20+3]= 400;        daa[5*20+4]= 10;         daa[6*20+0]= 499;        
	daa[6*20+1]= 152;        daa[6*20+2]= 1055;       daa[6*20+3]= 3691;       daa[6*20+4]= 10;         
	daa[6*20+5]= 3122;       daa[7*20+0]= 665;        daa[7*20+1]= 243;        daa[7*20+2]= 653;        
	daa[7*20+3]= 431;        daa[7*20+4]= 303;        daa[7*20+5]= 133;        daa[7*20+6]= 379;        
	daa[8*20+0]= 66;         daa[8*20+1]= 715;        daa[8*20+2]= 1405;       daa[8*20+3]= 331;        
	daa[8*20+4]= 441;        daa[8*20+5]= 1269;       daa[8*20+6]= 162;        daa[8*20+7]= 19;         
	daa[9*20+0]= 145;        daa[9*20+1]= 136;        daa[9*20+2]= 168;        daa[9*20+3]= 10;         
	daa[9*20+4]= 280;        daa[9*20+5]= 92;         daa[9*20+6]= 148;        daa[9*20+7]= 40;         
	daa[9*20+8]= 29;         daa[10*20+0]= 197;       daa[10*20+1]= 203;       daa[10*20+2]= 113;       
	daa[10*20+3]= 10;        daa[10*20+4]= 396;       daa[10*20+5]= 286;       daa[10*20+6]= 82;        
	daa[10*20+7]= 20;        daa[10*20+8]= 66;        daa[10*20+9]= 1745;      daa[11*20+0]= 236;       
	daa[11*20+1]= 4482;      daa[11*20+2]= 2430;      daa[11*20+3]= 412;       daa[11*20+4]= 48;        
	daa[11*20+5]= 3313;      daa[11*20+6]= 2629;      daa[11*20+7]= 263;       daa[11*20+8]= 305;       
	daa[11*20+9]= 345;       daa[11*20+10]= 218;      daa[12*20+0]= 185;       daa[12*20+1]= 125;       
	daa[12*20+2]= 61;        daa[12*20+3]= 47;        daa[12*20+4]= 159;       daa[12*20+5]= 202;       
	daa[12*20+6]= 113;       daa[12*20+7]= 21;        daa[12*20+8]= 10;        daa[12*20+9]= 1772;      
	daa[12*20+10]= 1351;     daa[12*20+11]= 193;      daa[13*20+0]= 68;        daa[13*20+1]= 53;        
	daa[13*20+2]= 97;        daa[13*20+3]= 22;        daa[13*20+4]= 726;       daa[13*20+5]= 10;        
	daa[13*20+6]= 145;       daa[13*20+7]= 25;        daa[13*20+8]= 127;       daa[13*20+9]= 454;       
	daa[13*20+10]= 1268;     daa[13*20+11]= 72;       daa[13*20+12]= 327;      daa[14*20+0]= 490;       
	daa[14*20+1]= 87;        daa[14*20+2]= 173;       daa[14*20+3]= 170;       daa[14*20+4]= 285;       
	daa[14*20+5]= 323;       daa[14*20+6]= 185;       daa[14*20+7]= 28;        daa[14*20+8]= 152;       
	daa[14*20+9]= 117;       daa[14*20+10]= 219;      daa[14*20+11]= 302;      daa[14*20+12]= 100;      
	daa[14*20+13]= 43;       daa[15*20+0]= 2440;      daa[15*20+1]= 385;       daa[15*20+2]= 2085;      
	daa[15*20+3]= 590;       daa[15*20+4]= 2331;      daa[15*20+5]= 396;       daa[15*20+6]= 568;       
	daa[15*20+7]= 691;       daa[15*20+8]= 303;       daa[15*20+9]= 216;       daa[15*20+10]= 516;      
	daa[15*20+11]= 868;      daa[15*20+12]= 93;       daa[15*20+13]= 487;      daa[15*20+14]= 1202;     
	daa[16*20+0]= 1340;      daa[16*20+1]= 314;       daa[16*20+2]= 1393;      daa[16*20+3]= 266;       
	daa[16*20+4]= 576;       daa[16*20+5]= 241;       daa[16*20+6]= 369;       daa[16*20+7]= 92;        
	daa[16*20+8]= 32;        daa[16*20+9]= 1040;      daa[16*20+10]= 156;      daa[16*20+11]= 918;      
	daa[16*20+12]= 645;      daa[16*20+13]= 148;      daa[16*20+14]= 260;      daa[16*20+15]= 2151;     
	daa[17*20+0]= 14;        daa[17*20+1]= 230;       daa[17*20+2]= 40;        daa[17*20+3]= 18;        
	daa[17*20+4]= 435;       daa[17*20+5]= 53;        daa[17*20+6]= 63;        daa[17*20+7]= 82;        
	daa[17*20+8]= 69;        daa[17*20+9]= 42;        daa[17*20+10]= 159;      daa[17*20+11]= 10;       
	daa[17*20+12]= 86;       daa[17*20+13]= 468;      daa[17*20+14]= 49;       daa[17*20+15]= 73;       
	daa[17*20+16]= 29;       daa[18*20+0]= 56;        daa[18*20+1]= 323;       daa[18*20+2]= 754;       
	daa[18*20+3]= 281;       daa[18*20+4]= 1466;      daa[18*20+5]= 391;       daa[18*20+6]= 142;       
	daa[18*20+7]= 10;        daa[18*20+8]= 1971;      daa[18*20+9]= 89;        daa[18*20+10]= 189;      
	daa[18*20+11]= 247;      daa[18*20+12]= 215;      daa[18*20+13]= 2370;     daa[18*20+14]= 97;       
	daa[18*20+15]= 522;      daa[18*20+16]= 71;       daa[18*20+17]= 346;      daa[19*20+0]= 968;       
	daa[19*20+1]= 92;        daa[19*20+2]= 83;        daa[19*20+3]= 75;        daa[19*20+4]= 592;       
	daa[19*20+5]= 54;        daa[19*20+6]= 200;       daa[19*20+7]= 91;        daa[19*20+8]= 25;        
	daa[19*20+9]= 4797;      daa[19*20+10]= 865;      daa[19*20+11]= 249;      daa[19*20+12]= 475;      
	daa[19*20+13]= 317;      daa[19*20+14]= 122;      daa[19*20+15]= 167;      daa[19*20+16]= 760;      
	daa[19*20+17]= 10;       daa[19*20+18]= 119;      
   
	f[0]= 0.076;            f[1]= 0.062;            f[2]= 0.041;            f[3]= 0.037;            
	f[4]= 0.009;            f[5]= 0.038;            f[6]= 0.049;            f[7]= 0.084;            
	f[8]= 0.025;            f[9]= 0.081;            f[10]= 0.101;           f[11]= 0.05;            
	f[12]= 0.022;           f[13]= 0.051;           f[14]= 0.043;           f[15]= 0.062;           
	f[16]= 0.054;           f[17]= 0.018;           f[18]= 0.031;           f[19]= 0.066; 
      }
      break;
    case VT:
      {
	daa[1*20+0]= 0.233108;   daa[2*20+0]= 0.199097;   daa[2*20+1]= 0.210797;   daa[3*20+0]= 0.265145;   
	daa[3*20+1]= 0.105191;   daa[3*20+2]= 0.883422;   daa[4*20+0]= 0.227333;   daa[4*20+1]= 0.031726;   
	daa[4*20+2]= 0.027495;   daa[4*20+3]= 0.010313;   daa[5*20+0]= 0.310084;   daa[5*20+1]= 0.493763;   
	daa[5*20+2]= 0.2757;     daa[5*20+3]= 0.205842;   daa[5*20+4]= 0.004315;   daa[6*20+0]= 0.567957;   
	daa[6*20+1]= 0.25524;    daa[6*20+2]= 0.270417;   daa[6*20+3]= 1.599461;   daa[6*20+4]= 0.005321;   
	daa[6*20+5]= 0.960976;   daa[7*20+0]= 0.876213;   daa[7*20+1]= 0.156945;   daa[7*20+2]= 0.362028;   
	daa[7*20+3]= 0.311718;   daa[7*20+4]= 0.050876;   daa[7*20+5]= 0.12866;    daa[7*20+6]= 0.250447;   
	daa[8*20+0]= 0.078692;   daa[8*20+1]= 0.213164;   daa[8*20+2]= 0.290006;   daa[8*20+3]= 0.134252;   
	daa[8*20+4]= 0.016695;   daa[8*20+5]= 0.315521;   daa[8*20+6]= 0.104458;   daa[8*20+7]= 0.058131;   
	daa[9*20+0]= 0.222972;   daa[9*20+1]= 0.08151;    daa[9*20+2]= 0.087225;   daa[9*20+3]= 0.01172;    
	daa[9*20+4]= 0.046398;   daa[9*20+5]= 0.054602;   daa[9*20+6]= 0.046589;   daa[9*20+7]= 0.051089;   
	daa[9*20+8]= 0.020039;   daa[10*20+0]= 0.42463;   daa[10*20+1]= 0.192364;  daa[10*20+2]= 0.069245;  
	daa[10*20+3]= 0.060863;  daa[10*20+4]= 0.091709;  daa[10*20+5]= 0.24353;   daa[10*20+6]= 0.151924;  
	daa[10*20+7]= 0.087056;  daa[10*20+8]= 0.103552;  daa[10*20+9]= 2.08989;   daa[11*20+0]= 0.393245;  
	daa[11*20+1]= 1.755838;  daa[11*20+2]= 0.50306;   daa[11*20+3]= 0.261101;  daa[11*20+4]= 0.004067;  
	daa[11*20+5]= 0.738208;  daa[11*20+6]= 0.88863;   daa[11*20+7]= 0.193243;  daa[11*20+8]= 0.153323;  
	daa[11*20+9]= 0.093181;  daa[11*20+10]= 0.201204; daa[12*20+0]= 0.21155;   daa[12*20+1]= 0.08793;   
	daa[12*20+2]= 0.05742;   daa[12*20+3]= 0.012182;  daa[12*20+4]= 0.02369;   daa[12*20+5]= 0.120801;  
	daa[12*20+6]= 0.058643;  daa[12*20+7]= 0.04656;   daa[12*20+8]= 0.021157;  daa[12*20+9]= 0.493845;  
	daa[12*20+10]= 1.105667; daa[12*20+11]= 0.096474; daa[13*20+0]= 0.116646;  daa[13*20+1]= 0.042569;  
	daa[13*20+2]= 0.039769;  daa[13*20+3]= 0.016577;  daa[13*20+4]= 0.051127;  daa[13*20+5]= 0.026235;  
	daa[13*20+6]= 0.028168;  daa[13*20+7]= 0.050143;  daa[13*20+8]= 0.079807;  daa[13*20+9]= 0.32102;   
	daa[13*20+10]= 0.946499; daa[13*20+11]= 0.038261; daa[13*20+12]= 0.173052; daa[14*20+0]= 0.399143;  
	daa[14*20+1]= 0.12848;   daa[14*20+2]= 0.083956;  daa[14*20+3]= 0.160063;  daa[14*20+4]= 0.011137;  
	daa[14*20+5]= 0.15657;   daa[14*20+6]= 0.205134;  daa[14*20+7]= 0.124492;  daa[14*20+8]= 0.078892;  
	daa[14*20+9]= 0.054797;  daa[14*20+10]= 0.169784; daa[14*20+11]= 0.212302; daa[14*20+12]= 0.010363; 
	daa[14*20+13]= 0.042564; daa[15*20+0]= 1.817198;  daa[15*20+1]= 0.292327;  daa[15*20+2]= 0.847049;  
	daa[15*20+3]= 0.461519;  daa[15*20+4]= 0.17527;   daa[15*20+5]= 0.358017;  daa[15*20+6]= 0.406035;  
	daa[15*20+7]= 0.612843;  daa[15*20+8]= 0.167406;  daa[15*20+9]= 0.081567;  daa[15*20+10]= 0.214977; 
	daa[15*20+11]= 0.400072; daa[15*20+12]= 0.090515; daa[15*20+13]= 0.138119; daa[15*20+14]= 0.430431; 
	daa[16*20+0]= 0.877877;  daa[16*20+1]= 0.204109;  daa[16*20+2]= 0.471268;  daa[16*20+3]= 0.178197;  
	daa[16*20+4]= 0.079511;  daa[16*20+5]= 0.248992;  daa[16*20+6]= 0.321028;  daa[16*20+7]= 0.136266;  
	daa[16*20+8]= 0.101117;  daa[16*20+9]= 0.376588;  daa[16*20+10]= 0.243227; daa[16*20+11]= 0.446646; 
	daa[16*20+12]= 0.184609; daa[16*20+13]= 0.08587;  daa[16*20+14]= 0.207143; daa[16*20+15]= 1.767766; 
	daa[17*20+0]= 0.030309;  daa[17*20+1]= 0.046417;  daa[17*20+2]= 0.010459;  daa[17*20+3]= 0.011393;  
	daa[17*20+4]= 0.007732;  daa[17*20+5]= 0.021248;  daa[17*20+6]= 0.018844;  daa[17*20+7]= 0.02399;   
	daa[17*20+8]= 0.020009;  daa[17*20+9]= 0.034954;  daa[17*20+10]= 0.083439; daa[17*20+11]= 0.023321; 
	daa[17*20+12]= 0.022019; daa[17*20+13]= 0.12805;  daa[17*20+14]= 0.014584; daa[17*20+15]= 0.035933; 
	daa[17*20+16]= 0.020437; daa[18*20+0]= 0.087061;  daa[18*20+1]= 0.09701;   daa[18*20+2]= 0.093268;  
	daa[18*20+3]= 0.051664;  daa[18*20+4]= 0.042823;  daa[18*20+5]= 0.062544;  daa[18*20+6]= 0.0552;    
	daa[18*20+7]= 0.037568;  daa[18*20+8]= 0.286027;  daa[18*20+9]= 0.086237;  daa[18*20+10]= 0.189842; 
	daa[18*20+11]= 0.068689; daa[18*20+12]= 0.073223; daa[18*20+13]= 0.898663; daa[18*20+14]= 0.032043; 
	daa[18*20+15]= 0.121979; daa[18*20+16]= 0.094617; daa[18*20+17]= 0.124746; daa[19*20+0]= 1.230985;  
	daa[19*20+1]= 0.113146;  daa[19*20+2]= 0.049824;  daa[19*20+3]= 0.048769;  daa[19*20+4]= 0.163831;  
	daa[19*20+5]= 0.112027;  daa[19*20+6]= 0.205868;  daa[19*20+7]= 0.082579;  daa[19*20+8]= 0.068575;  
	daa[19*20+9]= 3.65443;   daa[19*20+10]= 1.337571; daa[19*20+11]= 0.144587; daa[19*20+12]= 0.307309; 
	daa[19*20+13]= 0.247329; daa[19*20+14]= 0.129315; daa[19*20+15]= 0.1277;   daa[19*20+16]= 0.740372; 
	daa[19*20+17]= 0.022134; daa[19*20+18]= 0.125733; 
	   
	f[0]= 0.078837;         f[1]= 0.051238;         f[2]= 0.042313;         f[3]= 0.053066;         
	f[4]= 0.015175;         f[5]= 0.036713;         f[6]= 0.061924;         f[7]= 0.070852;         
	f[8]= 0.023082;         f[9]= 0.062056;         f[10]= 0.096371;        f[11]= 0.057324;        
	f[12]= 0.023771;        f[13]= 0.043296;        f[14]= 0.043911;        f[15]= 0.063403;        
	f[16]= 0.055897;        f[17]= 0.013272;        f[18]= 0.034399;        f[19]= 0.073101; 
      }
      break;
    case BLOSUM62:
      {
	daa[1*20+0]= 0.735790389698;  daa[2*20+0]= 0.485391055466;  daa[2*20+1]= 1.297446705134;  daa[3*20+0]= 0.543161820899;  
	daa[3*20+1]= 0.500964408555;  daa[3*20+2]= 3.180100048216;  daa[4*20+0]= 1.45999531047;   daa[4*20+1]= 0.227826574209;  
	daa[4*20+2]= 0.397358949897;  daa[4*20+3]= 0.240836614802;  daa[5*20+0]= 1.199705704602;  daa[5*20+1]= 3.020833610064;  
	daa[5*20+2]= 1.839216146992;  daa[5*20+3]= 1.190945703396;  daa[5*20+4]= 0.32980150463;   daa[6*20+0]= 1.1709490428;    
	daa[6*20+1]= 1.36057419042;   daa[6*20+2]= 1.24048850864;   daa[6*20+3]= 3.761625208368;  daa[6*20+4]= 0.140748891814;  
	daa[6*20+5]= 5.528919177928;  daa[7*20+0]= 1.95588357496;   daa[7*20+1]= 0.418763308518;  daa[7*20+2]= 1.355872344485;  
	daa[7*20+3]= 0.798473248968;  daa[7*20+4]= 0.418203192284;  daa[7*20+5]= 0.609846305383;  daa[7*20+6]= 0.423579992176;  
	daa[8*20+0]= 0.716241444998;  daa[8*20+1]= 1.456141166336;  daa[8*20+2]= 2.414501434208;  daa[8*20+3]= 0.778142664022;  
	daa[8*20+4]= 0.354058109831;  daa[8*20+5]= 2.43534113114;   daa[8*20+6]= 1.626891056982;  daa[8*20+7]= 0.539859124954;  
	daa[9*20+0]= 0.605899003687;  daa[9*20+1]= 0.232036445142;  daa[9*20+2]= 0.283017326278;  daa[9*20+3]= 0.418555732462;  
	daa[9*20+4]= 0.774894022794;  daa[9*20+5]= 0.236202451204;  daa[9*20+6]= 0.186848046932;  daa[9*20+7]= 0.189296292376;  
	daa[9*20+8]= 0.252718447885;  daa[10*20+0]= 0.800016530518; daa[10*20+1]= 0.622711669692; daa[10*20+2]= 0.211888159615; 
	daa[10*20+3]= 0.218131577594; daa[10*20+4]= 0.831842640142; daa[10*20+5]= 0.580737093181; daa[10*20+6]= 0.372625175087; 
	daa[10*20+7]= 0.217721159236; daa[10*20+8]= 0.348072209797; daa[10*20+9]= 3.890963773304; daa[11*20+0]= 1.295201266783; 
	daa[11*20+1]= 5.411115141489; daa[11*20+2]= 1.593137043457; daa[11*20+3]= 1.032447924952; daa[11*20+4]= 0.285078800906; 
	daa[11*20+5]= 3.945277674515; daa[11*20+6]= 2.802427151679; daa[11*20+7]= 0.752042440303; daa[11*20+8]= 1.022507035889; 
	daa[11*20+9]= 0.406193586642; daa[11*20+10]= 0.445570274261;daa[12*20+0]= 1.253758266664; daa[12*20+1]= 0.983692987457; 
	daa[12*20+2]= 0.648441278787; daa[12*20+3]= 0.222621897958; daa[12*20+4]= 0.76768882348;  daa[12*20+5]= 2.494896077113; 
	daa[12*20+6]= 0.55541539747;  daa[12*20+7]= 0.459436173579; daa[12*20+8]= 0.984311525359; daa[12*20+9]= 3.364797763104; 
	daa[12*20+10]= 6.030559379572;daa[12*20+11]= 1.073061184332;daa[13*20+0]= 0.492964679748; daa[13*20+1]= 0.371644693209; 
	daa[13*20+2]= 0.354861249223; daa[13*20+3]= 0.281730694207; daa[13*20+4]= 0.441337471187; daa[13*20+5]= 0.14435695975;  
	daa[13*20+6]= 0.291409084165; daa[13*20+7]= 0.368166464453; daa[13*20+8]= 0.714533703928; daa[13*20+9]= 1.517359325954; 
	daa[13*20+10]= 2.064839703237;daa[13*20+11]= 0.266924750511;daa[13*20+12]= 1.77385516883; daa[14*20+0]= 1.173275900924; 
	daa[14*20+1]= 0.448133661718; daa[14*20+2]= 0.494887043702; daa[14*20+3]= 0.730628272998; daa[14*20+4]= 0.356008498769; 
	daa[14*20+5]= 0.858570575674; daa[14*20+6]= 0.926563934846; daa[14*20+7]= 0.504086599527; daa[14*20+8]= 0.527007339151; 
	daa[14*20+9]= 0.388355409206; daa[14*20+10]= 0.374555687471;daa[14*20+11]= 1.047383450722;daa[14*20+12]= 0.454123625103;
	daa[14*20+13]= 0.233597909629;daa[15*20+0]= 4.325092687057; daa[15*20+1]= 1.12278310421;  daa[15*20+2]= 2.904101656456; 
	daa[15*20+3]= 1.582754142065; daa[15*20+4]= 1.197188415094; daa[15*20+5]= 1.934870924596; daa[15*20+6]= 1.769893238937; 
	daa[15*20+7]= 1.509326253224; daa[15*20+8]= 1.11702976291;  daa[15*20+9]= 0.35754441246;  daa[15*20+10]= 0.352969184527;
	daa[15*20+11]= 1.752165917819;daa[15*20+12]= 0.918723415746;daa[15*20+13]= 0.540027644824;daa[15*20+14]= 1.169129577716;
	daa[16*20+0]= 1.729178019485; daa[16*20+1]= 0.914665954563; daa[16*20+2]= 1.898173634533; daa[16*20+3]= 0.934187509431; 
	daa[16*20+4]= 1.119831358516; daa[16*20+5]= 1.277480294596; daa[16*20+6]= 1.071097236007; daa[16*20+7]= 0.641436011405; 
	daa[16*20+8]= 0.585407090225; daa[16*20+9]= 1.17909119726;  daa[16*20+10]= 0.915259857694;daa[16*20+11]= 1.303875200799;
	daa[16*20+12]= 1.488548053722;daa[16*20+13]= 0.488206118793;daa[16*20+14]= 1.005451683149;daa[16*20+15]= 5.15155629227; 
	daa[17*20+0]= 0.465839367725; daa[17*20+1]= 0.426382310122; daa[17*20+2]= 0.191482046247; daa[17*20+3]= 0.145345046279; 
	daa[17*20+4]= 0.527664418872; daa[17*20+5]= 0.758653808642; daa[17*20+6]= 0.407635648938; daa[17*20+7]= 0.508358924638; 
	daa[17*20+8]= 0.30124860078;  daa[17*20+9]= 0.34198578754;  daa[17*20+10]= 0.6914746346;  daa[17*20+11]= 0.332243040634;
	daa[17*20+12]= 0.888101098152;daa[17*20+13]= 2.074324893497;daa[17*20+14]= 0.252214830027;daa[17*20+15]= 0.387925622098;
	daa[17*20+16]= 0.513128126891;daa[18*20+0]= 0.718206697586; daa[18*20+1]= 0.720517441216; daa[18*20+2]= 0.538222519037; 
	daa[18*20+3]= 0.261422208965; daa[18*20+4]= 0.470237733696; daa[18*20+5]= 0.95898974285;  daa[18*20+6]= 0.596719300346; 
	daa[18*20+7]= 0.308055737035; daa[18*20+8]= 4.218953969389; daa[18*20+9]= 0.674617093228; daa[18*20+10]= 0.811245856323;
	daa[18*20+11]= 0.7179934869;  daa[18*20+12]= 0.951682162246;daa[18*20+13]= 6.747260430801;daa[18*20+14]= 0.369405319355;
	daa[18*20+15]= 0.796751520761;daa[18*20+16]= 0.801010243199;daa[18*20+17]= 4.054419006558;daa[19*20+0]= 2.187774522005; 
	daa[19*20+1]= 0.438388343772; daa[19*20+2]= 0.312858797993; daa[19*20+3]= 0.258129289418; daa[19*20+4]= 1.116352478606; 
	daa[19*20+5]= 0.530785790125; daa[19*20+6]= 0.524253846338; daa[19*20+7]= 0.25334079019;  daa[19*20+8]= 0.20155597175;  
	daa[19*20+9]= 8.311839405458; daa[19*20+10]= 2.231405688913;daa[19*20+11]= 0.498138475304;daa[19*20+12]= 2.575850755315;
	daa[19*20+13]= 0.838119610178;daa[19*20+14]= 0.496908410676;daa[19*20+15]= 0.561925457442;daa[19*20+16]= 2.253074051176;
	daa[19*20+17]= 0.266508731426;daa[19*20+18]= 1;             
	   
	f[0]= 0.074;                 f[1]= 0.052;                 f[2]= 0.045;                 f[3]= 0.054;                 
	f[4]= 0.025;                 f[5]= 0.034;                 f[6]= 0.054;                 f[7]= 0.074;                 
	f[8]= 0.026;                 f[9]= 0.068;                 f[10]= 0.099;                f[11]= 0.058;                
	f[12]= 0.025;                f[13]= 0.047;                f[14]= 0.039;                f[15]= 0.057;                
	f[16]= 0.051;                f[17]= 0.013;                f[18]= 0.032;                f[19]= 0.073; 
      }
      break;
    case MTMAM:
      {
	daa[1*20+0]= 32;              daa[2*20+0]= 2;    daa[2*20+1]= 4;               daa[3*20+0]= 11;
	daa[3*20+1]= 0;               daa[3*20+2]= 864;  daa[4*20+0]= 0;               daa[4*20+1]= 186;
	daa[4*20+2]= 0;               daa[4*20+3]= 0;    daa[5*20+0]= 0;               daa[5*20+1]= 246;
	daa[5*20+2]= 8;               daa[5*20+3]= 49;   daa[5*20+4]= 0;               daa[6*20+0]= 0;
	daa[6*20+1]= 0;               daa[6*20+2]= 0;    daa[6*20+3]= 569;             daa[6*20+4]= 0;
	daa[6*20+5]= 274;             daa[7*20+0]= 78;   daa[7*20+1]= 18;              daa[7*20+2]= 47;
	daa[7*20+3]= 79;              daa[7*20+4]= 0;    daa[7*20+5]= 0;               daa[7*20+6]= 22;
	daa[8*20+0]= 8;               daa[8*20+1]= 232;  daa[8*20+2]= 458;             daa[8*20+3]= 11;
	daa[8*20+4]= 305;             daa[8*20+5]= 550;  daa[8*20+6]= 22;              daa[8*20+7]= 0;
	daa[9*20+0]= 75;              daa[9*20+1]= 0;    daa[9*20+2]= 19;              daa[9*20+3]= 0;
	daa[9*20+4]= 41;              daa[9*20+5]= 0;    daa[9*20+6]= 0;               daa[9*20+7]= 0;
	daa[9*20+8]= 0;               daa[10*20+0]= 21;  daa[10*20+1]= 6;              daa[10*20+2]= 0;
	daa[10*20+3]= 0;              daa[10*20+4]= 27;  daa[10*20+5]= 20;             daa[10*20+6]= 0;
	daa[10*20+7]= 0;              daa[10*20+8]= 26;  daa[10*20+9]= 232;            daa[11*20+0]= 0;
	daa[11*20+1]= 50;             daa[11*20+2]= 408; daa[11*20+3]= 0;              daa[11*20+4]= 0;
	daa[11*20+5]= 242;            daa[11*20+6]= 215; daa[11*20+7]= 0;              daa[11*20+8]= 0;
	daa[11*20+9]= 6;              daa[11*20+10]= 4;  daa[12*20+0]= 76;             daa[12*20+1]= 0;
	daa[12*20+2]= 21;             daa[12*20+3]= 0;   daa[12*20+4]= 0;              daa[12*20+5]= 22;
	daa[12*20+6]= 0;              daa[12*20+7]= 0;   daa[12*20+8]= 0;              daa[12*20+9]= 378;
	daa[12*20+10]= 609;           daa[12*20+11]= 59; daa[13*20+0]= 0;              daa[13*20+1]= 0;
	daa[13*20+2]= 6;              daa[13*20+3]= 5;   daa[13*20+4]= 7;              daa[13*20+5]= 0;
	daa[13*20+6]= 0;              daa[13*20+7]= 0;   daa[13*20+8]= 0;              daa[13*20+9]= 57;
	daa[13*20+10]= 246;           daa[13*20+11]= 0;  daa[13*20+12]= 11;            daa[14*20+0]= 53;
	daa[14*20+1]= 9;              daa[14*20+2]= 33;  daa[14*20+3]= 2;              daa[14*20+4]= 0;
	daa[14*20+5]= 51;             daa[14*20+6]= 0;   daa[14*20+7]= 0;              daa[14*20+8]= 53;
	daa[14*20+9]= 5;              daa[14*20+10]= 43; daa[14*20+11]= 18;            daa[14*20+12]= 0;
	daa[14*20+13]= 17;            daa[15*20+0]= 342; daa[15*20+1]= 3;              daa[15*20+2]= 446;
	daa[15*20+3]= 16;             daa[15*20+4]= 347; daa[15*20+5]= 30;             daa[15*20+6]= 21;
	daa[15*20+7]= 112;            daa[15*20+8]= 20;  daa[15*20+9]= 0;              daa[15*20+10]= 74;
	daa[15*20+11]= 65;            daa[15*20+12]= 47; daa[15*20+13]= 90;            daa[15*20+14]= 202;
	daa[16*20+0]= 681;            daa[16*20+1]= 0;   daa[16*20+2]= 110;            daa[16*20+3]= 0;
	daa[16*20+4]= 114;            daa[16*20+5]= 0;   daa[16*20+6]= 4;              daa[16*20+7]= 0;
	daa[16*20+8]= 1;              daa[16*20+9]= 360; daa[16*20+10]= 34;            daa[16*20+11]= 50;
	daa[16*20+12]= 691;           daa[16*20+13]= 8;  daa[16*20+14]= 78;            daa[16*20+15]= 614;
	daa[17*20+0]= 5;              daa[17*20+1]= 16;  daa[17*20+2]= 6;              daa[17*20+3]= 0;
	daa[17*20+4]= 65;             daa[17*20+5]= 0;   daa[17*20+6]= 0;              daa[17*20+7]= 0;
	daa[17*20+8]= 0;              daa[17*20+9]= 0;   daa[17*20+10]= 12;            daa[17*20+11]= 0;
	daa[17*20+12]= 13;            daa[17*20+13]= 0;  daa[17*20+14]= 7;             daa[17*20+15]= 17;
	daa[17*20+16]= 0;             daa[18*20+0]= 0;   daa[18*20+1]= 0;              daa[18*20+2]= 156;
	daa[18*20+3]= 0;              daa[18*20+4]= 530; daa[18*20+5]= 54;             daa[18*20+6]= 0;
	daa[18*20+7]= 1;              daa[18*20+8]= 1525;daa[18*20+9]= 16;             daa[18*20+10]= 25;
	daa[18*20+11]= 67;            daa[18*20+12]= 0;  daa[18*20+13]= 682;           daa[18*20+14]= 8;
	daa[18*20+15]= 107;           daa[18*20+16]= 0;  daa[18*20+17]= 14;            daa[19*20+0]= 398;
	daa[19*20+1]= 0;              daa[19*20+2]= 0;   daa[19*20+3]= 10;             daa[19*20+4]= 0;
	daa[19*20+5]= 33;             daa[19*20+6]= 20;  daa[19*20+7]= 5;              daa[19*20+8]= 0;
	daa[19*20+9]= 2220;           daa[19*20+10]= 100;daa[19*20+11]= 0;             daa[19*20+12]= 832;
	daa[19*20+13]= 6;             daa[19*20+14]= 0;  daa[19*20+15]= 0;             daa[19*20+16]= 237;
	daa[19*20+17]= 0;             daa[19*20+18]= 0;       
    
	f[0]= 0.0692;  f[1]=  0.0184;  f[2]= 0.04;    f[3]= 0.0186;
	f[4]= 0.0065;  f[5]=  0.0238;  f[6]= 0.0236;  f[7]= 0.0557;
	f[8]= 0.0277;  f[9]=  0.0905;  f[10]=0.1675;  f[11]= 0.0221;
	f[12]=0.0561;  f[13]= 0.0611;  f[14]=0.0536;  f[15]= 0.0725;
	f[16]=0.087;   f[17]= 0.0293;  f[18]=0.034;   f[19]= 0.0428;
      }
      break;
    case LG:
      {
      //      readExternalRateMatrix (initialRates, f, "LG.PAML.txt");
	double aaMut[20][20], a[20][20];

	aaMut[1][0]=0.425093; 
	aaMut[2][0]=0.276818; aaMut[2][1]=0.751878; 
	aaMut[3][0]=0.395144; aaMut[3][1]=0.123954; aaMut[3][2]=5.076149; 
	aaMut[4][0]=2.489084; aaMut[4][1]=0.534551; aaMut[4][2]=0.528768; aaMut[4][3]=0.062556; 
	aaMut[5][0]=0.969894; aaMut[5][1]=2.807908; aaMut[5][2]=1.695752; aaMut[5][3]=0.523386; aaMut[5][4]=0.084808; 
	aaMut[6][0]=1.038545; aaMut[6][1]=0.363970; aaMut[6][2]=0.541712; aaMut[6][3]=5.243870; aaMut[6][4]=0.003499; aaMut[6][5]=4.128591; 
	aaMut[7][0]=2.066040; aaMut[7][1]=0.390192; aaMut[7][2]=1.437645; aaMut[7][3]=0.844926; aaMut[7][4]=0.569265; aaMut[7][5]=0.267959; aaMut[7][6]=0.348847; 
	aaMut[8][0]=0.358858; aaMut[8][1]=2.426601; aaMut[8][2]=4.509238; aaMut[8][3]=0.927114; aaMut[8][4]=0.640543; aaMut[8][5]=4.813505; aaMut[8][6]=0.423881; aaMut[8][7]=0.311484; 
	aaMut[9][0]=0.149830; aaMut[9][1]=0.126991; aaMut[9][2]=0.191503; aaMut[9][3]=0.010690; aaMut[9][4]=0.320627; aaMut[9][5]=0.072854; aaMut[9][6]=0.044265; aaMut[9][7]=0.008705; aaMut[9][8]=0.108882; 
	aaMut[10][0]=0.395337; aaMut[10][1]=0.301848; aaMut[10][2]=0.068427; aaMut[10][3]=0.015076; aaMut[10][4]=0.594007; aaMut[10][5]=0.582457; aaMut[10][6]=0.069673; aaMut[10][7]=0.044261; aaMut[10][8]=0.366317; aaMut[10][9]=4.145067; 
	aaMut[11][0]=0.536518; aaMut[11][1]=6.326067; aaMut[11][2]=2.145078; aaMut[11][3]=0.282959; aaMut[11][4]=0.013266; aaMut[11][5]=3.234294; aaMut[11][6]=1.807177; aaMut[11][7]=0.296636; aaMut[11][8]=0.697264; aaMut[11][9]=0.159069; aaMut[11][10]=0.137500; 
	aaMut[12][0]=1.124035; aaMut[12][1]=0.484133; aaMut[12][2]=0.371004; aaMut[12][3]=0.025548; aaMut[12][4]=0.893680; aaMut[12][5]=1.672569; aaMut[12][6]=0.173735; aaMut[12][7]=0.139538; aaMut[12][8]=0.442472; aaMut[12][9]=4.273607; aaMut[12][10]=6.312358; aaMut[12][11]=0.656604; 
	aaMut[13][0]=0.253701; aaMut[13][1]=0.052722; aaMut[13][2]=0.089525; aaMut[13][3]=0.017416; aaMut[13][4]=1.105251; aaMut[13][5]=0.035855; aaMut[13][6]=0.018811; aaMut[13][7]=0.089586; aaMut[13][8]=0.682139; aaMut[13][9]=1.112727; aaMut[13][10]=2.592692; aaMut[13][11]=0.023918; aaMut[13][12]=1.798853; 
	aaMut[14][0]=1.177651; aaMut[14][1]=0.332533; aaMut[14][2]=0.161787; aaMut[14][3]=0.394456; aaMut[14][4]=0.075382; aaMut[14][5]=0.624294; aaMut[14][6]=0.419409; aaMut[14][7]=0.196961; aaMut[14][8]=0.508851; aaMut[14][9]=0.078281; aaMut[14][10]=0.249060; aaMut[14][11]=0.390322; aaMut[14][12]=0.099849; aaMut[14][13]=0.094464; 
	aaMut[15][0]=4.727182; aaMut[15][1]=0.858151; aaMut[15][2]=4.008358; aaMut[15][3]=1.240275; aaMut[15][4]=2.784478; aaMut[15][5]=1.223828; aaMut[15][6]=0.611973; aaMut[15][7]=1.739990; aaMut[15][8]=0.990012; aaMut[15][9]=0.064105; aaMut[15][10]=0.182287; aaMut[15][11]=0.748683; aaMut[15][12]=0.346960; aaMut[15][13]=0.361819; aaMut[15][14]=1.338132; 
	aaMut[16][0]=2.139501; aaMut[16][1]=0.578987; aaMut[16][2]=2.000679; aaMut[16][3]=0.425860; aaMut[16][4]=1.143480; aaMut[16][5]=1.080136; aaMut[16][6]=0.604545; aaMut[16][7]=0.129836; aaMut[16][8]=0.584262; aaMut[16][9]=1.033739; aaMut[16][10]=0.302936; aaMut[16][11]=1.136863; aaMut[16][12]=2.020366; aaMut[16][13]=0.165001; aaMut[16][14]=0.571468; aaMut[16][15]=6.472279; 
	aaMut[17][0]=0.180717; aaMut[17][1]=0.593607; aaMut[17][2]=0.045376; aaMut[17][3]=0.029890; aaMut[17][4]=0.670128; aaMut[17][5]=0.236199; aaMut[17][6]=0.077852; aaMut[17][7]=0.268491; aaMut[17][8]=0.597054; aaMut[17][9]=0.111660; aaMut[17][10]=0.619632; aaMut[17][11]=0.049906; aaMut[17][12]=0.696175; aaMut[17][13]=2.457121; aaMut[17][14]=0.095131; aaMut[17][15]=0.248862; aaMut[17][16]=0.140825; 
	aaMut[18][0]=0.218959; aaMut[18][1]=0.314440; aaMut[18][2]=0.612025; aaMut[18][3]=0.135107; aaMut[18][4]=1.165532; aaMut[18][5]=0.257336; aaMut[18][6]=0.120037; aaMut[18][7]=0.054679; aaMut[18][8]=5.306834; aaMut[18][9]=0.232523; aaMut[18][10]=0.299648; aaMut[18][11]=0.131932; aaMut[18][12]=0.481306; aaMut[18][13]=7.803902; aaMut[18][14]=0.089613; aaMut[18][15]=0.400547; aaMut[18][16]=0.245841; aaMut[18][17]=3.151815; 
	aaMut[19][0]=2.547870; aaMut[19][1]=0.170887; aaMut[19][2]=0.083688; aaMut[19][3]=0.037967; aaMut[19][4]=1.959291; aaMut[19][5]=0.210332; aaMut[19][6]=0.245034; aaMut[19][7]=0.076701; aaMut[19][8]=0.119013; aaMut[19][9]=10.649107; aaMut[19][10]=1.702745; aaMut[19][11]=0.185202; aaMut[19][12]=1.898718; aaMut[19][13]=0.654683; aaMut[19][14]=0.296501; aaMut[19][15]=0.098369; aaMut[19][16]=2.188158; aaMut[19][17]=0.189510; aaMut[19][18]=0.249313; 
	
	// from triangular exchange matrix to square exchange  matrix
	for (i=0; i<20; i++) {
	  for (j=0; j<=i; j++) 
	    a[i][j] = aaMut[i][j];
	  
	  for (j=i+1; j<20; j++)
	    a[i][j] = aaMut[j][i];  
	}
	
	r = 0;
	for(i = 0; i < 19; i++)
	  {      
	    for(j = i + 1; j < 20; j++) 
	      initialRates[r++] = a[i][j];
	  }
	
	f[0]=0.079066; f[1]=0.055941; f[2]=0.041977; f[3]=0.053052; f[4]=0.012937; f[5]=0.040767; f[6]=0.071586; f[7]=0.057337; f[8]=0.022355; f[9]=0.062157; f[10]=0.099081; f[11]=0.064600; f[12]=0.022951; f[13]=0.042302; f[14]=0.044040; f[15]=0.061197; f[16]=0.053287; f[17]=0.012066; f[18]=0.034155; f[19]=0.069147; 
	
	for (i=0; i<20; i++) protModelFreq[i] = f[i];

	return (0);    /* for LG model no further codes needed */
      }

    case GTR:
      printf("this function should not be called by GTR model for proteins\n");
      exit(-1);
    default: 
      printf("FATAL ERROR not defined\n");
      exit(-1);
    }

  for (i=0; i<20; i++) protModelFreq[i] = f[i];


  for (i=0; i<20; i++)  
    for (j=0; j<i; j++)  
      daa[j*20+i] = daa[i*20+j];

  max = 0;
  for(i = 0; i < 19; i++)
    for(j = i + 1; j < 20; j++)
      {
	q[i][j] = temp = daa[i * 20 + j];
	if(temp > max) 
	  max = temp;
      }

  r = 0;
  for(i = 0; i < 19; i++)
    {      
      for(j = i + 1; j < 20; j++)
	{ 
	  initialRates[r] = q[i][j];
	  r++;
	}
    }      
       
  return 0;
}


/* 4 cF profiles of site frequencies from PCA analyses of 7758 sites of 21 datasets: Wang, Li, Susko and Roger BMC evol. Biol. 2008 */
/* use -@ on qmmraxml command line to invoke cF4 */
void initCFvectors(int model, double f[20])
{

  switch (model) {
  case 1:
    {f[0]=0.02549352; f[1]=0.01296012; f[2]=0.005545202; f[3]=0.006005566; f[4]=0.01002193; f[5]=0.01112289; f[6]=0.008811948; f[7]=0.001796161; f[8]=0.004312188; f[9]=0.2108274; f[10]=0.2730413; f[11]=0.01335451; f[12]=0.07862202; f[13]=0.03859909; f[14]=0.005058205; f[15]=0.008209453; f[16]=0.03210019; f[17]=0.002668138; f[18]=0.01379098; f[19]=0.2376598;}
    break;
    
  case 2:
    {f[0]=0.09596966; f[1]=0.008786096; f[2]=0.02805857; f[3]=0.01880183; f[4]=0.005026264; f[5]=0.006454635; f[6]=0.01582725; f[7]=0.7215719; f[8]=0.003379354; f[9]=0.002257725; f[10]=0.003013483; f[11]=0.01343441; f[12]=0.001511657; f[13]=0.002107865; f[14]=0.006751404; f[15]=0.04798539; f[16]=0.01141559; f[17]=0.000523736; f[18]=0.002188483; f[19]=0.004934972;}    
    break;

  case 3:
    {f[0]=0.01726065; f[1]=0.005467988; f[2]=0.01092937; f[3]=0.3627871; f[4]=0.001046402; f[5]=0.01984758; f[6]=0.5149206; f[7]=0.004145081; f[8]=0.002563289; f[9]=0.002955213; f[10]=0.005286931; f[11]=0.01558693; f[12]=0.002693098; f[13]=0.002075771; f[14]=0.003006167; f[15]=0.01263069; f[16]=0.01082144; f[17]=0.000253451; f[18]=0.001144787; f[19]=0.004573568;}
    break;

  case 4:
    {f[0]=0.1263139; f[1]=0.09564027; f[2]=0.07050061; f[3]=0.03316681; f[4]=0.02095119; f[5]=0.05473468; f[6]=0.02790523; f[7]=0.009007538; f[8]=0.03441334; f[9]=0.005855319; f[10]=0.008061884; f[11]=0.1078084; f[12]=0.009019514; f[13]=0.05018693; f[14]=0.07948; f[15]=0.09447839; f[16]=0.09258897; f[17]=0.01390669; f[18]=0.05367769; f[19]=0.01230413;}
    break;

  default:
    printf("CF model out of range: must be 4 CF vectors in total. \n");
    exit(-1);
  }  
}


/* the 9 aa profiles from Sjölander et al. Comput Appl Biosci 12:327–345, 1996
use -B on qmmraxml command line to invoke cF9 */
void init9DirichiletPriors_CFvectors(int model, double f[20])
{
  switch (model) {
  case 1:
    {f[0]=0.229256; f[1]=0.019999; f[2]=0.040908; f[3]=0.014887; f[4]=0.033751; f[5]=0.017501; f[6]=0.013903; f[7]=0.111732; f[8]=0.010495; f[9]=0.019141; f[10]=0.026025; f[11]=0.017243; f[12]=0.012972; f[13]=0.012085; f[14]=0.045571; f[15]=0.183075; f[16]=0.124699; f[17]=0.003183; f[18]=0.008149; f[19]=0.055425;}
    break;

  case 2:
    {f[0]=0.015832; f[1]=0.013915; f[2]=0.021089; f[3]=0.008660; f[4]=0.007597; f[5]=0.016972; f[6]=0.008027; f[7]=0.012108; f[8]=0.056199; f[9]=0.026057; f[10]=0.068974; f[11]=0.010268; f[12]=0.016251; f[13]=0.284439; f[14]=0.009652; f[15]=0.021504; f[16]=0.013389; f[17]=0.052934; f[18]=0.309509; f[19]=0.026626;}
    break;

  case 3:
    {f[0]=0.084248; f[1]=0.069854; f[2]=0.066258; f[3]=0.065778; f[4]=0.006820; f[5]=0.079654; f[6]=0.114665; f[7]=0.038881; f[8]=0.032252; f[9]=0.021897; f[10]=0.037111; f[11]=0.114370; f[12]=0.017805; f[13]=0.013109; f[14]=0.026232; f[15]=0.087541; f[16]=0.066861; f[17]=0.004428; f[18]=0.018170; f[19]=0.034069;}
    break;

  case 4:
    {f[0]=0.033700; f[1]=0.243481; f[2]=0.038614; f[3]=0.009359; f[4]=0.005352; f[5]=0.088900; f[6]=0.045477; f[7]=0.023080; f[8]=0.036994; f[9]=0.015825; f[10]=0.034733; f[11]=0.277042; f[12]=0.013568; f[13]=0.006324; f[14]=0.018094; f[15]=0.035424; f[16]=0.034394; f[17]=0.005407; f[18]=0.013800; f[19]=0.020434;}
    break;

  case 5:
    {f[0]=0.019751; f[1]=0.006972; f[2]=0.002944; f[3]=0.002696; f[4]=0.007109; f[5]=0.009514; f[6]=0.004909; f[7]=0.003747; f[8]=0.003448; f[9]=0.143985; f[10]=0.480270; f[11]=0.005213; f[12]=0.101003; f[13]=0.073811; f[14]=0.006257; f[15]=0.005790; f[16]=0.017203; f[17]=0.006124; f[18]=0.012718; f[19]=0.086537;}
    break;

  case 6:
    {f[0]=0.045015; f[1]=0.004944; f[2]=0.005665; f[3]=0.004834; f[4]=0.014555; f[5]=0.004432; f[6]=0.007079; f[7]=0.006719; f[8]=0.001912; f[9]=0.310289; f[10]=0.111307; f[11]=0.006648; f[12]=0.029519; f[13]=0.020161; f[14]=0.005877; f[15]=0.010722; f[16]=0.034395; f[17]=0.001703; f[18]=0.006519; f[19]=0.367706;}
    break;

  case 7:
    {f[0]=0.052921; f[1]=0.021895; f[2]=0.106377; f[3]=0.219275; f[4]=0.002682; f[5]=0.062308; f[6]=0.196959; f[7]=0.059951; f[8]=0.028185; f[9]=0.008473; f[10]=0.015719; f[11]=0.053382; f[12]=0.005685; f[13]=0.006128; f[14]=0.028322; f[15]=0.067648; f[16]=0.037259; f[17]=0.001820; f[18]=0.010612; f[19]=0.014399;}
    break;

  case 8:
    {f[0]=0.090658; f[1]=0.028798; f[2]=0.023772; f[3]=0.012523; f[4]=0.022979; f[5]=0.025397; f[6]=0.023198; f[7]=0.028110; f[8]=0.020121; f[9]=0.110318; f[10]=0.140476; f[11]=0.028870; f[12]=0.055453; f[13]=0.056990; f[14]=0.019542; f[15]=0.055934; f[16]=0.071873; f[17]=0.012337; f[18]=0.039973; f[19]=0.132677;}
    break;

  case 9:
    {f[0]=0.052191; f[1]=0.066161; f[2]=0.041789; f[3]=0.067558; f[4]=0.040593; f[5]=0.036482; f[6]=0.061518; f[7]=0.170161; f[8]=0.036653; f[9]=0.021950; f[10]=0.060201; f[11]=0.050442; f[12]=0.014804; f[13]=0.034854; f[14]=0.091005; f[15]=0.031879; f[16]=0.037085; f[17]=0.027859; f[18]=0.026995; f[19]=0.029819;}
    break;

  default:
    printf("CF model out of range: must be 9 CF vectors in total. \n");
    exit(-1);
  }  
}


/* unpublished: 15 cF profiles of amino acid site frequencies from a Self-Organizing Map analyses of 35584 sites of a eukaryotic dataset: 143 proteins 48 taxa (Hampl et al 2009)
use -A on qmmraxml command line to invoke cF15: not recommendated */
void initCF15vectors(int model, double f[20])
{
  switch (model) {
  case 1:
{f[0]=0.010917; f[1]=0.002196; f[2]=0.001222; f[3]=0.000453; f[4]=0.006329; f[5]=0.002617; f[6]=0.001691; f[7]=0.000871; f[8]=0.001228; f[9]=0.276871; f[10]=0.343318; f[11]=0.002592; f[12]=0.020796; f[13]=0.007050; f[14]=0.001677; f[15]=0.002412; f[16]=0.008791; f[17]=0.000325; f[18]=0.001964; f[19]=0.306679; }
    break;

  case 2:
{f[0]=0.014082; f[1]=0.007539; f[2]=0.005451; f[3]=0.001486; f[4]=0.004796; f[5]=0.013248; f[6]=0.003917; f[7]=0.001998; f[8]=0.004817; f[9]=0.057114; f[10]=0.240600; f[11]=0.008328; f[12]=0.509895; f[13]=0.048147; f[14]=0.002431; f[15]=0.006641; f[16]=0.009896; f[17]=0.004322; f[18]=0.018017; f[19]=0.037274; }
    break;

  case 3:
{f[0]=0.007689; f[1]=0.004220; f[2]=0.005035; f[3]=0.001863; f[4]=0.004105; f[5]=0.003755; f[6]=0.002887; f[7]=0.002352; f[8]=0.016487; f[9]=0.009135; f[10]=0.025572; f[11]=0.003228; f[12]=0.008677; f[13]=0.450037; f[14]=0.001330; f[15]=0.005757; f[16]=0.004297; f[17]=0.077179; f[18]=0.357307; f[19]=0.009086; }
    break;

  case 4:
{f[0]=0.078363; f[1]=0.024194; f[2]=0.012544; f[3]=0.007409; f[4]=0.045261; f[5]=0.031354; f[6]=0.023159; f[7]=0.006201; f[8]=0.011435; f[9]=0.098116; f[10]=0.154330; f[11]=0.029222; f[12]=0.040849; f[13]=0.013393; f[14]=0.015322; f[15]=0.028185; f[16]=0.100964; f[17]=0.001960; f[18]=0.009871; f[19]=0.267866; }
    break;

  case 5:
{f[0]=0.044148; f[1]=0.035506; f[2]=0.018196; f[3]=0.048329; f[4]=0.012268; f[5]=0.047610; f[6]=0.147207; f[7]=0.016534; f[8]=0.081862; f[9]=0.038307; f[10]=0.119683; f[11]=0.034561; f[12]=0.074356; f[13]=0.037116; f[14]=0.023047; f[15]=0.019343; f[16]=0.041212; f[17]=0.017031; f[18]=0.072621; f[19]=0.071062; }
    break;

  case 6:
{f[0]=0.017615; f[1]=0.004898; f[2]=0.017914; f[3]=0.428572; f[4]=0.000644; f[5]=0.017943; f[6]=0.446437; f[7]=0.009406; f[8]=0.003727; f[9]=0.001690; f[10]=0.002936; f[11]=0.012081; f[12]=0.001604; f[13]=0.001046; f[14]=0.005563; f[15]=0.013208; f[16]=0.008868; f[17]=0.000123; f[18]=0.001582; f[19]=0.004143; }
    break;

  case 7:
{f[0]=0.051886; f[1]=0.006684; f[2]=0.014690; f[3]=0.004642; f[4]=0.154475; f[5]=0.010303; f[6]=0.007160; f[7]=0.007857; f[8]=0.002928; f[9]=0.016378; f[10]=0.012545; f[11]=0.010869; f[12]=0.008660; f[13]=0.002508; f[14]=0.005536; f[15]=0.083052; f[16]=0.551046; f[17]=0.000273; f[18]=0.002105; f[19]=0.046401; }
    break;

  case 8:
{f[0]=0.084096; f[1]=0.031873; f[2]=0.060820; f[3]=0.050609; f[4]=0.011938; f[5]=0.060376; f[6]=0.090531; f[7]=0.022218; f[8]=0.048340; f[9]=0.020272; f[10]=0.044257; f[11]=0.057832; f[12]=0.032551; f[13]=0.015231; f[14]=0.010272; f[15]=0.134551; f[16]=0.147275; f[17]=0.002218; f[18]=0.017369; f[19]=0.057370; }
    break;

  case 9:
{f[0]=0.041117; f[1]=0.021692; f[2]=0.059922; f[3]=0.076041; f[4]=0.002599; f[5]=0.059254; f[6]=0.113204; f[7]=0.019200; f[8]=0.418939; f[9]=0.006304; f[10]=0.013061; f[11]=0.046129; f[12]=0.006247; f[13]=0.010673; f[14]=0.008834; f[15]=0.037450; f[16]=0.026020; f[17]=0.001787; f[18]=0.020439; f[19]=0.011089; }
    break;

  case 10:
{f[0]=0.137562; f[1]=0.008921; f[2]=0.024801; f[3]=0.012271; f[4]=0.010660; f[5]=0.014423; f[6]=0.011806; f[7]=0.035374; f[8]=0.005713; f[9]=0.003134; f[10]=0.004780; f[11]=0.012458; f[12]=0.003569; f[13]=0.001880; f[14]=0.016147; f[15]=0.632435; f[16]=0.052454; f[17]=0.000411; f[18]=0.002313; f[19]=0.008888; }
    break;

  case 11:
{f[0]=0.111361; f[1]=0.068305; f[2]=0.100832; f[3]=0.031994; f[4]=0.004353; f[5]=0.105563; f[6]=0.054833; f[7]=0.044804; f[8]=0.034175; f[9]=0.008702; f[10]=0.018725; f[11]=0.126293; f[12]=0.011740; f[13]=0.004865; f[14]=0.017378; f[15]=0.155401; f[16]=0.068093; f[17]=0.000751; f[18]=0.006294; f[19]=0.025539; }
    break;

  case 12:
{f[0]=0.023524; f[1]=0.026438; f[2]=0.379248; f[3]=0.023410; f[4]=0.002848; f[5]=0.337980; f[6]=0.029147; f[7]=0.014703; f[8]=0.023984; f[9]=0.004415; f[10]=0.010291; f[11]=0.046514; f[12]=0.008245; f[13]=0.002286; f[14]=0.004516; f[15]=0.029212; f[16]=0.021485; f[17]=0.000240; f[18]=0.003655; f[19]=0.007861; }
    break;

  case 13:
{f[0]=0.315670; f[1]=0.002986; f[2]=0.006854; f[3]=0.004816; f[4]=0.006638; f[5]=0.005025; f[6]=0.005530; f[7]=0.384142; f[8]=0.002001; f[9]=0.002438; f[10]=0.003423; f[11]=0.005423; f[12]=0.001609; f[13]=0.001123; f[14]=0.205314; f[15]=0.028338; f[16]=0.008539; f[17]=0.000250; f[18]=0.001042; f[19]=0.008838; }
    break;

  case 14:
{f[0]=0.205884; f[1]=0.074902; f[2]=0.036811; f[3]=0.025754; f[4]=0.002789; f[5]=0.042325; f[6]=0.050083; f[7]=0.118733; f[8]=0.009885; f[9]=0.009223; f[10]=0.022335; f[11]=0.175725; f[12]=0.006363; f[13]=0.006395; f[14]=0.087364; f[15]=0.067471; f[16]=0.033916; f[17]=0.000524; f[18]=0.003291; f[19]=0.020227; }
    break;

  case 15:
{f[0]=0.014996; f[1]=0.424476; f[2]=0.012941; f[3]=0.004260; f[4]=0.001740; f[5]=0.021248; f[6]=0.011438; f[7]=0.006351; f[8]=0.008159; f[9]=0.004283; f[10]=0.009006; f[11]=0.433364; f[12]=0.004641; f[13]=0.001797; f[14]=0.004455; f[15]=0.012759; f[16]=0.013629; f[17]=0.000435; f[18]=0.002237; f[19]=0.007785; }
    break;

  default:
    printf("CF15 model out of range: must be 15 CF vectors in total \n");
    exit(-1);
  }  

  int i;
  for (i=0;i<20;i++) {if (f[i] < FREQ_MIN) f[i] = FREQ_MIN;}
  
}


/* CAT C20 profiles of the equlibirium amino acid site frequencies from analyzing HSSP dataset, described in Quang, Gascuel & Lartillot: Bioinformatics 24: 2317-2323 (2008)
use -D on qmmraxml command line to invoke cF20
Note: if use -C on qmmraxml command line it will invoke CAT20 model with F81 rate matrix  */
void initCAT_C20vectors(int model, double f[20])
{
  switch (model) {
  case 0:
{f[0]=0.08624130; f[1]=0.01719440; f[2]=0.07912930; f[3]=0.03299090; f[4]=0.01305050; f[5]=0.01690470; f[6]=0.01845270; f[7]=0.03669050; f[8]=0.01080130; f[9]=0.00979071; f[10]=0.01128260; f[11]=0.02201950; f[12]=0.00878215; f[13]=0.00441553; f[14]=0.01892730; f[15]=0.31781500; f[16]=0.27117000; f[17]=0.00153173; f[18]=0.00483429; f[19]=0.01797530; }
    break;

  case 1:
{f[0]=0.20355800; f[1]=0.00509808; f[2]=0.00770524; f[3]=0.00316561; f[4]=0.03486670; f[5]=0.00640441; f[6]=0.00708594; f[7]=0.01952360; f[8]=0.00243920; f[9]=0.11525700; f[10]=0.07897770; f[11]=0.00423808; f[12]=0.03091870; f[13]=0.01124290; f[14]=0.01641890; f[15]=0.04967770; f[16]=0.11189500; f[17]=0.00177626; f[18]=0.00484482; f[19]=0.28490600; }
    break;

  case 2:
{f[0]=0.02115470; f[1]=0.00149462; f[2]=0.00127550; f[3]=0.00054929; f[4]=0.00481886; f[5]=0.00123288; f[6]=0.00145396; f[7]=0.00114309; f[8]=0.00113464; f[9]=0.39284600; f[10]=0.12506400; f[11]=0.00135799; f[12]=0.02097890; f[13]=0.01282520; f[14]=0.00202472; f[15]=0.00262407; f[16]=0.01719140; f[17]=0.00115911; f[18]=0.00360280; f[19]=0.38606800; }
    break;

  case 3:
{f[0]=0.03769040; f[1]=0.28852000; f[2]=0.03654110; f[3]=0.01094690; f[4]=0.00640738; f[5]=0.08935640; f[6]=0.03583650; f[7]=0.01911070; f[8]=0.03295140; f[9]=0.01017120; f[10]=0.02374960; f[11]=0.28976300; f[12]=0.00965289; f[13]=0.00363498; f[14]=0.01053370; f[15]=0.03563140; f[16]=0.03559270; f[17]=0.00279252; f[18]=0.00665572; f[19]=0.01446220; }
    break;

  case 4:
{f[0]=0.00845978; f[1]=0.00535899; f[2]=0.00725259; f[3]=0.00244879; f[4]=0.00849090; f[5]=0.00427815; f[6]=0.00250555; f[7]=0.00242771; f[8]=0.04332140; f[9]=0.00977130; f[10]=0.03805070; f[11]=0.00267410; f[12]=0.00807248; f[13]=0.34204600; f[14]=0.00214187; f[15]=0.00804189; f[16]=0.00553221; f[17]=0.04948400; f[18]=0.43750000; f[19]=0.01214100; }
    break;

  case 5:
{f[0]=0.17599000; f[1]=0.02904290; f[2]=0.03328460; f[3]=0.13012600; f[4]=0.00175587; f[5]=0.07071840; f[6]=0.21821700; f[7]=0.04095350; f[8]=0.01307080; f[9]=0.00856221; f[10]=0.01595310; f[11]=0.05429460; f[12]=0.00540458; f[13]=0.00252770; f[14]=0.03710200; f[15]=0.07934810; f[16]=0.05400830; f[17]=0.00105921; f[18]=0.00362591; f[19]=0.02495530; }
    break;

  case 6:
{f[0]=0.16344000; f[1]=0.01955410; f[2]=0.04387020; f[3]=0.03742730; f[4]=0.00886599; f[5]=0.01375550; f[6]=0.02206120; f[7]=0.52967200; f[8]=0.00900061; f[9]=0.00175694; f[10]=0.00611563; f[11]=0.01671180; f[12]=0.00293908; f[13]=0.00306413; f[14]=0.01264580; f[15]=0.08293430; f[16]=0.01428360; f[17]=0.00286407; f[18]=0.00323983; f[19]=0.00579857; }
    break;

  case 7:
{f[0]=0.09174690; f[1]=0.02658530; f[2]=0.02906990; f[3]=0.01338190; f[4]=0.02840150; f[5]=0.02550850; f[6]=0.01968760; f[7]=0.02498990; f[8]=0.04497660; f[9]=0.05835560; f[10]=0.11550100; f[11]=0.01649160; f[12]=0.03959950; f[13]=0.09984790; f[14]=0.02099160; f[15]=0.07364830; f[16]=0.06615180; f[17]=0.02464640; f[18]=0.09723270; f[19]=0.08318560; }
    break;

  case 8:
{f[0]=0.06467010; f[1]=0.09880160; f[2]=0.02289070; f[3]=0.01687340; f[4]=0.00771176; f[5]=0.09964150; f[6]=0.05449780; f[7]=0.01488940; f[8]=0.03138520; f[9]=0.05059830; f[10]=0.18442800; f[11]=0.09079310; f[12]=0.07748400; f[13]=0.02191480; f[14]=0.01050040; f[15]=0.03211960; f[16]=0.04117660; f[17]=0.00843030; f[18]=0.02061060; f[19]=0.05058240; }
    break;

  case 9:
{f[0]=0.01359940; f[1]=0.00434084; f[2]=0.00184694; f[3]=0.00079517; f[4]=0.01000900; f[5]=0.00464208; f[6]=0.00180118; f[7]=0.00267946; f[8]=0.00724019; f[9]=0.08140270; f[10]=0.36614200; f[11]=0.00251581; f[12]=0.07349650; f[13]=0.26409700; f[14]=0.00389941; f[15]=0.00436688; f[16]=0.00752485; f[17]=0.02615650; f[18]=0.06609710; f[19]=0.05734730; }
    break;

  case 10:
{f[0]=0.14780400; f[1]=0.08428450; f[2]=0.07266300; f[3]=0.05347430; f[4]=0.00488258; f[5]=0.07571660; f[6]=0.07272460; f[7]=0.09077260; f[8]=0.02622890; f[9]=0.00357811; f[10]=0.01267770; f[11]=0.10516600; f[12]=0.00596218; f[13]=0.00299039; f[14]=0.01565580; f[15]=0.14599000; f[16]=0.06348770; f[17]=0.00159285; f[18]=0.00507607; f[19]=0.00927198; }
    break;

  case 11:
{f[0]=0.01863770; f[1]=0.00420552; f[2]=0.00198652; f[3]=0.00083297; f[4]=0.00549689; f[5]=0.00658901; f[6]=0.00202485; f[7]=0.00217135; f[8]=0.00236660; f[9]=0.20208100; f[10]=0.43703800; f[11]=0.00291207; f[12]=0.12418600; f[13]=0.03853830; f[14]=0.00406723; f[15]=0.00461774; f[16]=0.01499040; f[17]=0.00268717; f[18]=0.00563241; f[19]=0.11893800; }
    break;

  case 12:
{f[0]=0.04776240; f[1]=0.05057430; f[2]=0.02095740; f[3]=0.01413490; f[4]=0.00757917; f[5]=0.04292970; f[6]=0.04626880; f[7]=0.00523279; f[8]=0.01653520; f[9]=0.17415000; f[10]=0.11212500; f[11]=0.05775750; f[12]=0.03302880; f[13]=0.01306910; f[14]=0.01243750; f[15]=0.02649890; f[16]=0.09517550; f[17]=0.00316605; f[18]=0.01124660; f[19]=0.20937000; }
    break;

  case 13:
{f[0]=0.41641900; f[1]=0.00561008; f[2]=0.00917014; f[3]=0.00451317; f[4]=0.04069380; f[5]=0.00613205; f[6]=0.00632298; f[7]=0.09461850; f[8]=0.00310574; f[9]=0.00764432; f[10]=0.00998854; f[11]=0.00389418; f[12]=0.00693232; f[13]=0.00484384; f[14]=0.01878410; f[15]=0.23607700; f[16]=0.07462750; f[17]=0.00121726; f[18]=0.00348258; f[19]=0.04592250; }
    break;

  case 14:
{f[0]=0.04022960; f[1]=0.07352030; f[2]=0.10366500; f[3]=0.03655240; f[4]=0.01247830; f[5]=0.08265580; f[6]=0.03721970; f[7]=0.02336180; f[8]=0.21083100; f[9]=0.00934787; f[10]=0.03605610; f[11]=0.04824110; f[12]=0.01002900; f[13]=0.04590950; f[14]=0.00985040; f[15]=0.05333830; f[16]=0.03102090; f[17]=0.01400770; f[18]=0.10643800; f[19]=0.01524800; }
    break;

  case 15:
{f[0]=0.03234530; f[1]=0.02362830; f[2]=0.25204500; f[3]=0.24315000; f[4]=0.00359763; f[5]=0.03308310; f[6]=0.07102740; f[7]=0.10160700; f[8]=0.03662250; f[9]=0.00314108; f[10]=0.00519805; f[11]=0.04701290; f[12]=0.00240287; f[13]=0.00244293; f[14]=0.00948378; f[15]=0.08483550; f[16]=0.03590830; f[17]=0.00087309; f[18]=0.00672477; f[19]=0.00487046; }
    break;

  case 16:
{f[0]=0.14762600; f[1]=0.03345070; f[2]=0.02119730; f[3]=0.04030520; f[4]=0.00323272; f[5]=0.03715540; f[6]=0.05768930; f[7]=0.03308510; f[8]=0.01463930; f[9]=0.01082670; f[10]=0.02562010; f[11]=0.04513510; f[12]=0.00586514; f[13]=0.00471772; f[14]=0.34737100; f[15]=0.08920650; f[16]=0.04858990; f[17]=0.00163587; f[18]=0.00441772; f[19]=0.02823360; }
    break;

  case 17:
{f[0]=0.10314500; f[1]=0.07177480; f[2]=0.04351720; f[3]=0.03864020; f[4]=0.00617625; f[5]=0.07866030; f[6]=0.09233690; f[7]=0.02023380; f[8]=0.02467620; f[9]=0.03769040; f[10]=0.03762840; f[11]=0.09216990; f[12]=0.01618830; f[13]=0.00676664; f[14]=0.01283020; f[15]=0.09514500; f[16]=0.13785700; f[17]=0.00221447; f[18]=0.00830416; f[19]=0.07404540; }
    break;

  case 18:
{f[0]=0.08375430; f[1]=0.08993830; f[2]=0.05188110; f[3]=0.08048710; f[4]=0.00207351; f[5]=0.14565000; f[6]=0.19477600; f[7]=0.02290300; f[8]=0.02684590; f[9]=0.00740798; f[10]=0.01902500; f[11]=0.14592900; f[12]=0.00673952; f[13]=0.00230634; f[14]=0.00856160; f[15]=0.04557400; f[16]=0.04510810; f[17]=0.00107713; f[18]=0.00493253; f[19]=0.01503030; }
    break;

  case 19:
{f[0]=0.05787360; f[1]=0.01383140; f[2]=0.04914220; f[3]=0.29467400; f[4]=0.00111308; f[5]=0.05982500; f[6]=0.34021000; f[7]=0.02939110; f[8]=0.01398170; f[9]=0.00305257; f[10]=0.00626119; f[11]=0.03633650; f[12]=0.00272960; f[13]=0.00170349; f[14]=0.01561060; f[15]=0.03580450; f[16]=0.02499420; f[17]=0.00086643; f[18]=0.00383130; f[19]=0.00876742; }
    break;

  default:
    printf("CAT_C20 model out of range: must be 20 C20 vectors in total \n");
    exit(-1);
  }

}


void initUserProtMat(tree *tr, int model, double f[20])
{
  int i, j, r, modelNo, count;
  double *initialRates = &(tr->initialRates[model * tr->numberOfRates]);
  FILE *myFile;
  double aaMut[20][20], a[20][20];
  char *pch, buf[256];

  myFile = fopen(userModelsFileName,"r");

  if (!myFile) {
    if (processID == 0)
      printf("Could not open user-defined models file: %s\n", userModelsFileName);
    errorExit(-1);
  }

  fgets (buf, 256, myFile);    // skip 1st line

  // read exchange rate matrix
  modelNo = 0;
  i = 1; j = 0;     // the diagnal values not provideed in the exchange matrix, so the matrix start at i=1
  while (fgets (buf, 256, myFile) != NULL) {
    if (buf[0] == '>') {
      modelNo ++;
      if (modelNo < model) {
	do {
	  fgets (buf, 256, myFile);
	  if (buf[0] == '>') modelNo++;
	} while (modelNo < model); 
      }
      
      for (count = 1; count < 20; count ++) { 
	fgets (buf, 256, myFile);
	pch = strtok(buf, " ");
	while (pch != NULL) {
	  if (*pch == '\n') break;
	  aaMut[i][j++] = atof(pch);
	  pch = strtok(NULL, " ");
	}
	i ++; j = 0;
      }
      
      // get aa freq
      fgets (buf, 256, myFile);   // skip the blank line
      fgets (buf, 256, myFile);  
      i = 0;
      pch = strtok(buf, " ,\t");
      while (pch != NULL){
	if (*pch == '\n') break;
	f[i++] = atof(pch);
	pch = strtok (NULL, " ,\t");
      }
    }
    break;
  }   // finish reading user rate matrix and aa freq


  // from triangular exchange matrix to square exchange  matrix
  for (i=0; i<20; i++) {
    for (j=0; j<=i; j++) 
      a[i][j] = aaMut[i][j];
     
    for (j=i+1; j<20; j++)
      a[i][j] = aaMut[j][i];  
  }

  r = 0;
  for(i = 0; i < 19; i++)
    {      
      for(j = i + 1; j < 20; j++) 
	initialRates[r++] = a[i][j];
    }

  fclose (myFile);
}


void readAAfreqWithoutSelection (double pi_j0[20])
{
  FILE *aafreqFile;
  char buf[100];
  int i=0;
  float a;

  aafreqFile = fopen(AAfreq_noselect_FileName, "r");
  
  if (!aafreqFile)
    {
      if(processID == 0)
	printf( "Could not open user-defined AA freq without-selection file: %s\n", AAfreq_noselect_FileName);

      errorExit(-1);
    }   

  while (fgets(buf, 100, aafreqFile)) 
    {
      sscanf(buf, "%f", &a);
      pi_j0[i] = a;
      if (pi_j0[i] == 0.0) pi_j0[i] = FREQ_MIN;
      i++;
      if (i>20) {
	printf("too many data! check the amino acid freq. file\n");
	errorExit(-1);
      }
    }
}


void resetMatrix (double M[20][20])
{
  int i,j;

  for (i = 0; i < 20; i++) 
    for (j = 0; j < 20; j++) 
      M[i][j] = 0.;
}


void rescaleQmatrix (double Q[20][20], double f[20], int model)
{
  int i, j;
  double offDiag = 0.0, rowSum, max;

  for (i=0; i<20; i++) {
    rowSum = 0.0; 
    for (j=0; j<20; j++) {
      if (i != j) rowSum += Q[i][j];
    }
    Q[i][i] = -1.0 * rowSum;
    offDiag += rowSum * f[i];
  }
  
  /*  printf("\noffdiag=%f\n", offDiag);   */

  // scale the matrix so that the expected number of substitutions per unit time = 1
  // i.e., let Sum_i Sum_j!=i pi_i * qij = 1

  for (i=0; i<20; i++) 
    for (j=0; j<20; j++) 
      Q[i][j] /= offDiag;
  
  /*
  double total=0, sumtemp=0;
  if (model==0) printf("\nvarify the q matrix\n");
  printf("model=%d\n", model);
  for (i=0; i<20; i++) {
    total += f[i] * Q[i][i];
    sumtemp = 0.;
    for (j=0; j<20; j++) sumtemp += Q[i][j];
    
    printf("row sum =%f\n", sumtemp);
  }
  printf("sum pi Qii=%f\n", total);
  */

  /* check stationarity: pi_i * Q_ij = pi_j * Q_ji */
  /*  
  printf("\n\ncheck stationarity:\n");
  double temp; 
  max = 0.;
  for (i=0; i<20; i++) { 
    for (j=0; j<20; j++) {
      temp = abs(f[i]*Q[i][j] - f[j]*Q[j][i]);
      if (temp > max) max = temp;
    }
  }
  printf("\nmax diff in Q stationarity check is %f.\n\n", max);
  */
}


void printQmatrix(int model, tree *tr, double Qmatrix[20][20])
{
  int i,j;

  char QmatrixFileName[1024];
  FILE *QmatrixFile;
  strcpy(QmatrixFileName, "RAxML_proteinQmatrix.dat");

  if (model==0) QmatrixFile = fopen(QmatrixFileName, "w");
  else QmatrixFile = fopen(QmatrixFileName, "a");
      
  if (model <= tr->numberOfUserModels) fprintf(QmatrixFile, "model%d ", model);
  else fprintf(QmatrixFile, "model%ds ", model-tr->numberOfUserModels);


  for (i=0; i<20; i++) {
    for (j=0; j<20; j++) {
      fprintf(QmatrixFile, "%8.6f ", Qmatrix[i][j]);
    }
  }
  fprintf(QmatrixFile, "\n");
  
  fclose (QmatrixFile);
  if (model == tr->NumberOfModels + 2 * tr->numberOfUserModels) {
    printf("\nPrinted protein substitution matrix Q to file %s\n\n", QmatrixFileName);
  }
}

int getQmatrix(int model, tree *tr, double Q[20][20], double f[20])
{
  int i, j, r;
  double rates[190], aaMut[20][20];
      
  for(i = 0; i < tr->numberOfRates; i++)
    rates[i] = tr->initialRates[model * tr->numberOfRates  + i];
  
  resetMatrix (aaMut);
  
  r = 0;
  for(i = 0; i < 19; i++) {
    for(j = i + 1; j < 20; j++) {
      aaMut[i][j] = aaMut[j][i] = rates[r++];
    }
  }
  
  /*    define Q matrix   */
  for (i=0; i<20; i++)
    for (j=0; j<20; j++)
      Q[i][j] = aaMut[i][j] * f[j];

  /* rescale the Q matrix */
  rescaleQmatrix (Q, f, model);

  //  printQmatrix(model, tr, Q);
}


void selQmatrix(double *initialRates, int model, tree *tr, double Q[20][20])
{
  int i,j,k,l;
  double f[20],f_0[20], F_c[20], F_JTT[20], diff;
  double Q_JTT[20][20], Q_0[20][20];

  readAAfreqWithoutSelection (f_0);
  for (i=0; i<20; i++) f[i] = tr->frequencies[model * 20 + i];

  resetMatrix (Q_JTT);
  resetMatrix (Q_0);
  resetMatrix (Q);

  /* Q_JTT */
  getQmatrix (model, tr, Q_JTT, protModelFreq);

  /* amino acid relative fitness */
  for (i = 0; i < 20; i++) {
    F_JTT[i] = log(protModelFreq[i]/f_0[i]);
    F_c[i] = log(f[i]/f_0[i]);
  }
  
  /* Q_0  */
  for (i=0; i<20; i++)
    for (j=0; j<20; j++) {
      if (i != j) {
	diff = F_JTT[j]-F_JTT[i];
	if (diff == 0) {
	  Q_0[i][j] = Q_JTT[i][j];
	  printf("\nstrange case: %d\t%d\n", i, j);
	}
	else Q_0[i][j] = Q_JTT[i][j] * (1-exp(-diff))/diff;
      }
    }

  rescaleQmatrix (Q_0, f_0, model);
  //  printQmatrix(model, tr, Q_0);

  /* Q */
  for (i=0; i<20; i++)
    for (j=0; j<20; j++) {
      if (i != j) {
	diff = F_c[j]-F_c[i];
	if (diff == 0) {
	  Q[i][j] = Q_0[i][j];
	  printf("\nstrange case: %d\t%d\n", i, j);
	}
	else Q[i][j] = Q_0[i][j] * diff/(1-exp(-diff));
      }
    }

  rescaleQmatrix (Q, f, model);
  //  printQmatrix(model, tr, Q);
}


void initReversibleGTR(rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef, int model)
{ 
  if(adef->model == M_PROTCAT || adef->model == M_PROTGAMMA)
    {
      double r[20][20], a[20][20], *initialRates = &(tr->initialRates[model * 190]), 
	f[20], e[20], d[20];
      int i, j, k, m, jj, code, l;
      double invfreq[20], EIGN[20], EIGV[20][20]; 
      double *eptr;         
      double Qsel[20][20], Ssel[20][20];   // Q and S matricies under selection
      
      if(adef->useMultipleModel)
	{
	  if(adef->proteinMatrix == GTR)
	    {
	      printf("FATAL ERROR no GTR for AA with multiple models\n");
	      exit(-1);
	    }
	  else
	    {	     
	      initProtMat(tr, model, f, adef, adef->protModels[model]);
	      
	      if(adef->protFreqs[model])
		{	     	      
		  for(l = 0; l < 20; l++)		
		    f[l] = tr->frequencies[model * 20 + l];			       
		}
	      else
		{	     
		  for(l = 0; l < 20; l++)		
		    tr->frequencies[model * 20 + l] = f[l];
		}
	    }
	}
      else if (adef->useMultipleUserModels)
	{
	  // for -m defined rate matrix
	  if (model == 0) {
	    if(adef->proteinMatrix == GTR) {
	      for(l = 0; l < 20; l++)
		f[l] = tr->frequencies[model * 20 + l];
	    }
	    else { 
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      // for -m defined rate matrix; prot freq can be empirical or from the input data 
	      if(adef->protEmpiricalFreqs) {	     	      
		for(l = 0; l < 20; l++)		
		  f[l] = tr->frequencies[model * 20 + l];			       
		  
	      }
	      else {
		for(l = 0; l < 20; l++) 	
		  tr->frequencies[model * 20 + l] = f[l];
	      }
	    }
	  }   /*  end model == 0 */
	  else {
	   // for user defined rate matrices
	    initUserProtMat(tr, model, f);
	    for(l = 0; l < 20; l++) 
	      tr->frequencies[model * 20 + l] = f[l];
	  }
	}
      else if (adef->useCFmodel) {
	  // for -m defined rate matrix
	  if (model == 0) {
	    if(adef->proteinMatrix == GTR) {
	      for(l = 0; l < 20; l++)
		f[l] = tr->frequencies[model * 20 + l];
	    }
	    else { 
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      // for -m defined rate matrix; prot freq can be empirical or from the input data 
	      if(adef->protEmpiricalFreqs) {	     	      
		for(l = 0; l < 20; l++)		
		  f[l] = tr->frequencies[model * 20 + l];			       
		  
	      }
	      else {
		for(l = 0; l < 20; l++) 	
		  tr->frequencies[model * 20 + l] = f[l];
	      }
	    }
	  }   /*  end model == 0 */
	  else {
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      initCFvectors(model, f);
	      for(l = 0; l < 20; l++)		
		tr->frequencies[model * 20 + l] = f[l];			       
	  }
	  
      }

      else if (adef->useCF9model) {
	  // for -m defined rate matrix
	  if (model == 0) {
	    if(adef->proteinMatrix == GTR) {
	      for(l = 0; l < 20; l++)
		f[l] = tr->frequencies[model * 20 + l];
	    }
	    else { 
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      // for -m defined rate matrix; prot freq can be empirical or from the input data 
	      if(adef->protEmpiricalFreqs) {	     	      
		for(l = 0; l < 20; l++)		
		  f[l] = tr->frequencies[model * 20 + l];			       
		  
	      }
	      else {
		for(l = 0; l < 20; l++) 	
		  tr->frequencies[model * 20 + l] = f[l];
	      }
	    }
	  }   /*  end model == 0 */
	  else {
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      init9DirichiletPriors_CFvectors (model, f);
	      for(l = 0; l < 20; l++)		
		tr->frequencies[model * 20 + l] = f[l];			       
	  }
	  
      }

      else if (adef->useCF15model) {
	  // for -m defined rate matrix
	  if (model == 0) {
	    if(adef->proteinMatrix == GTR) {
	      for(l = 0; l < 20; l++)
		f[l] = tr->frequencies[model * 20 + l];
	    }
	    else { 
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      // for -m defined rate matrix; prot freq can be empirical or from the input data 
	      if(adef->protEmpiricalFreqs) {	     	      
		for(l = 0; l < 20; l++)		
		  f[l] = tr->frequencies[model * 20 + l];			       
		  
	      }
	      else {
		for(l = 0; l < 20; l++) 	
		  tr->frequencies[model * 20 + l] = f[l];
	      }
	    }
	  }   /*  end model == 0 */
	  else {
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      initCF15vectors(model, f);
	      for(l = 0; l < 20; l++)		
		tr->frequencies[model * 20 + l] = f[l];			       
	  }
	  
      }
      else if (adef->useCAT_C20Fmodel) { /* modified CAT_C20 and proportional model with an extra F component */
	  // for -m defined rate matrix
	  if (model == 0) {
	    initProtMat(tr, model, f, adef, adef->proteinMatrix);
	    for(l = 0; l < 20; l++)		
	      f[l] = tr->frequencies[model * 20 + l];			       
	  }
	  else {
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      initCAT_C20vectors(model-1, f);
	      for(l = 0; l < 20; l++)		
		tr->frequencies[model * 20 + l] = f[l];			       
	  }
	  
      }
       else if (adef->useCAT_C20model) {
	/* for all mixture components, use F81 rate matrix (all exchange rates = 1.0 as in F81/proportional model */
	initProtMat(tr, model, f, adef, adef->proteinMatrix);

	initCAT_C20vectors(model, f);  /* will overwrite the aa freq from the dataset */

	for(l = 0; l < 20; l++)		
	  tr->frequencies[model * 20 + l] = f[l];			       
      }
	  
      else
	{
	  if(adef->proteinMatrix == GTR)
	    {
	      for(l = 0; l < 20; l++)
		f[l] = tr->frequencies[model * 20 + l];
	    }
	  else
	    {	  
	      initProtMat(tr, model, f, adef, adef->proteinMatrix);
	      if(adef->protEmpiricalFreqs)
		{	     	      
		  for(l = 0; l < 20; l++)		
		    f[l] = tr->frequencies[model * 20 + l];			       
		}
	      else
		{	     
		  for(l = 0; l < 20; l++)	
		    tr->frequencies[model * 20 + l] = f[l];
		}
	    }
	}
                 

      i = 0;
      
      for(j = 0; j < 19; j++)
	for (k = j+1; k < 20; k++)      	  
	  r[j][k] = initialRates[i++];     
            
      for (j = 0; j < 20; j++) 
	{
	  r[j][j] = 0.0;
	  for (k = 0; k < j; k++)
	    r[j][k] = r[k][j];
	}           

      /* TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/

      tr->fracchanges[model] = 0;
      tr->fracchange = 0;

      for (j = 0; j< 20; j++)
	for (k = 0; k< 20; k++)
	  tr->fracchanges[model] += f[j] * r[j][k] * f[k];
      
      //      printf("\nfracchange[%d] = %f\n", model, tr->fracchanges[model]);

      i = 0;
      for (j = 0; j< 19; j++)
	for (k = j+1; k< 20; k++)
	  initialRates[i++] /= tr->fracchanges[model];  // normalize (i.e. scale) the initial rates by tr->fracchanges of the model, which is devided by the sum of the diagonal of the Q matrix for the model, so that the expected rate of sub. is the rate of changes per site and both hard-coded rate matrix and external input matrices can be used simultaneiusly, and, tr->fracchange will be 1.0. So the following line is commented
      
      /*      if(tr->NumberOfModels == 1 && !(adef->useMultipleUserModels) ) 
      	tr->fracchange = tr->fracchanges[model];
      */


      /* change Q matrix under selection model */
      if (adef->useSelectionModel && model == 0) {   /* Ed idea 10/03/2013: only F component under selection */
	selQmatrix(initialRates, model, tr, Qsel);   /* rates under selection models are scaled in the function */

	/* get the S exchange matrix corresponding to the Q matrix under selection */
	for (i=0; i<20; i++) {
	  for (j=0; j<20; j++) {
	    if (i != j) Ssel[i][j] = Qsel[i][j] /f[j];
	    else Ssel[i][j] = 0;
	  }
	}
      
	l = 0;
	for(i = 0; i < 19; i++)      
	  for(j = i + 1; j < 20; j++)
	    initialRates[l++] = Ssel[i][j];
      }


      /* TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/

      m = 0;

      for(i=0; i< 20; i++) 
	a[i][i] = 0;
      
      for(i=0; i < 20; i++) 
	{
	  for(j=i+1;  j < 20 ; j++) 
	    {
	      double factor =  initialRates[m++];
	      a[i][j] = a[j][i] = factor * sqrt( f[i] * f[j]);
	      a[i][i] -= factor * f[j];
	      a[j][j] -= factor * f[i];
	    }
	}
           
      tred2((double *)a,20,20,d,e);       
      tqli(d, e, 20 , 20, (double *)a);   
      
      for(i=0; i<20; i++)     
	for(j=0; j<20; j++)       
	  a[i][j] *= sqrt(f[j]);
      
      for (i=0; i<20; i++)
	{	  
	  if (d[i] > -1e-8) 
	    {	      
	      if (i != 0) 
		{		    
		  double tmp = d[i], sum=0;
		  d[i] = d[0];
		  d[0] = tmp;
		  for (j=0; j < 20; j++) 
		    {
		      tmp = a[i][j];
		      a[i][j] = a[0][j];
		      sum += (a[0][j] = tmp);
		    }
		  for (j=0; j < 20; j++) 
		    a[0][j] /= sum;
		}
	      break;
	    }
	}
  
      for (i=0; i< 20; i++) 
	{
	  EIGN[i] = -d[i];
	  
	  for (j=0; j<20; j++)
	    EIGV[i][j] = a[j][i];
	  invfreq[i] = 1 / EIGV[i][0]; 
	}                                    

      for(l = 1; l < 20; l++) {
	tr->EIGN[model * 19 + (l - 1)] = EIGN[l]; 
      }

      eptr = &(tr->EV[model * 400]);
      
      for(i = 0; i < 20; i++)		  
	for(j = 0; j < 20; j++)      	    
	  *eptr++ = EIGV[i][j];	             
	    	
      for(i = 0; i < 20; i++)
	for(j = 1; j < 20; j++)
	  tr->EI[model * 380 + i * 19 + (j - 1)] = tr->EV[model * 400 + i * 20 + j] * invfreq[i];
      
      for(i=0; i < 23; i++)
	{     
	  for(j = 0; j < 20; j++)
	    tr->protTip[model * 23 + i].v[j] = 0.0; 
	  
	  if(i < 20)
	    {	  
	      for (j = 0; j < 20; j++) 	    
		tr->protTip[model * 23 + i].v[j] += EIGV[i][j];	  
	    }
	  else
	    {
	      if(i == 20)
		{
		  for (j = 0; j < 20; j++) 	    
		    {
		      tr->protTip[model * 23 + i].v[j] += EIGV[2][j] + EIGV[3][j];
		    }
		}
	      else
		{
		  if(i == 21)
		    {
		      for (j = 0; j < 20; j++) 	    
			{
			  tr->protTip[model * 23 + i].v[j] += EIGV[5][j] + EIGV[6][j];
			}
		    }
		  else
		    {
		      if(i == 22)
			{
			  for (j = 0; j < 20; j++) 	    
			    {
			      for(k = 0; k < 20; k++)
				{
				  tr->protTip[model * 23 + i].v[j] += EIGV[k][j];	  			      
				}
			    }
			}
		    }
		}
	    }
	}
    }
  else
    {
      double r[4][4], a[4][4], *initialRates = &(tr->initialRates[model * 5]), 
	f[4], e[4], d[4];
      int i, j, k, m, jj, code;
      double invfreq[4], EIGN[4], EIGV[4][4]; 
      double *eptr;
    
      f[0] = tr->frequencies[model * 4];
      f[1] = tr->frequencies[model * 4 + 1];
      f[2] = tr->frequencies[model * 4 + 2];
      f[3] = tr->frequencies[model * 4 + 3];
      
      i = 0;
      for (j = 0; j < 2; j++)
	for (k = j+1; k< 4; k++)	  
	  r[j][k] = initialRates[i++];
      r[2][3] = 1.0;
      for (j = 0; j < 4; j++) 
	{
	  r[j][j] = 0;
	  for (k = 0; k < j; k++)
	    r[j][k] = r[k][j];
	}
      
      
      /* TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/
      tr->fracchanges[model] = 0;
      tr->fracchange = 0;
      
      for (j = 0; j<4; j++)
	for (k = 0; k<4; k++)
	  tr->fracchanges[model] += f[j] * r[j][k] * f[k];
      
      if(tr->NumberOfModels == 1)    
	tr->fracchange = tr->fracchanges[model];
      
      /*TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/
      m = 0;
      
      for(i=0; i<4; i++) 
	a[i][i] = 0;            

      for(i=0; i<4; i++) 
	{
	  for(j=i+1; j<4; j++) 
	    {
	      double factor = ((m>=5) ? 1.0 : initialRates[m++]);
	      a[i][j] = a[j][i] = factor * sqrt(f[i] * f[j]);
	      a[i][i] -= factor * f[j];
	      a[j][j] -= factor * f[i];	     
	    }
	}           
     
     
      tred2((double *)a,4,4,d,e);       
      tqli(d, e, 4 , 4, (double *)a);
     
     

      for(i=0; i<4; i++)     
	for(j=0; j<4; j++)       
	  a[i][j] *= sqrt(f[j]);
      
      for (i=0; i<4; i++)
	{	  
	  if (d[i] > -1e-8) 
	    {	      
	      if (i != 0)
		{		    
		  double tmp = d[i], sum=0;
		  d[i] = d[0];
		  d[0] = tmp;
		  for (j=0; j < 4; j++) 
		    {
		      tmp = a[i][j];
		      a[i][j] = a[0][j];
		      sum += (a[0][j] = tmp);
		    }
		  for (j=0; j < 4; j++) 
		    a[0][j] /= sum;
		}
	      break;
	    }
	}
  
      for (i=0; i< 4; i++) 
	{
	  EIGN[i] = -d[i];
	  
	  for (j=0; j<4; j++)
	    EIGV[i][j] = a[j][i];
	  invfreq[i] = 1 / EIGV[i][0]; 
	}                   
                 
      tr->EIGN[model * 3] = EIGN[1];
      tr->EIGN[model * 3 + 1] = EIGN[2];
      tr->EIGN[model * 3 + 2] = EIGN[3];        

      eptr = &(tr->EV[model * 16]);

      for(i = 0; i < 4; i++)       
	for(j = 0; j < 4; j++) 	  	  
	  *eptr++ = EIGV[i][j];	    	   
      
      tr->EI[model * 12]     = tr->EV[model * 16 + 1] * invfreq[0];
      tr->EI[model * 12 + 1] = tr->EV[model * 16 + 2] * invfreq[0];
      tr->EI[model * 12 + 2] = tr->EV[model * 16 + 3] * invfreq[0];
      
      tr->EI[model * 12 + 3] = tr->EV[model * 16 + 5] * invfreq[1];
      tr->EI[model * 12 + 4] = tr->EV[model * 16 + 6] * invfreq[1];
      tr->EI[model * 12 + 5] = tr->EV[model * 16 + 7] * invfreq[1];
      
      tr->EI[model * 12 + 6] = tr->EV[model * 16 + 9]  * invfreq[2];
      tr->EI[model * 12 + 7] = tr->EV[model * 16 + 10] * invfreq[2];
      tr->EI[model * 12 + 8] = tr->EV[model * 16 + 11] * invfreq[2];
      
      tr->EI[model * 12 + 9]  = tr->EV[model * 16 + 13] * invfreq[3];
      tr->EI[model * 12 + 10] = tr->EV[model * 16 + 14] * invfreq[3];
      tr->EI[model * 12 + 11] = tr->EV[model * 16 + 15] * invfreq[3];
           
      for(i=0; i < 16; i++)
	{
	  code = i;
	  tr->gtrTip[model * 16 + i].a = 0;
	  tr->gtrTip[model * 16 + i].c = 0;
	  tr->gtrTip[model * 16 + i].g = 0;
	  tr->gtrTip[model * 16 + i].t = 0;
	  if(i > 0)
	    {		      
	      for (j = 0; j < 4; j++) 
		{	    
		  if ((code >> j) & 1) 
		    {
		      int jj = "0123"[j] - '0';				 		     
		      tr->gtrTip[model * 16 + i].a += EIGV[jj][0];
		      tr->gtrTip[model * 16 + i].c += EIGV[jj][1];
		      tr->gtrTip[model * 16 + i].g += EIGV[jj][2];
		      tr->gtrTip[model * 16 + i].t += EIGV[jj][3];
		    }	     		  
		}	    
	    }     
	}
    }
  
}


double LnGamma (double alpha)
{
/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
  double x, f, z, result;

  x = alpha;
  f = 0.0;
  
  if ( x < 7.0) 
     {
       f = 1.0;  
       z = alpha - 1.0;
      
       while ((z = z + 1.0) < 7.0)  
	 {	  
	   f *= z;
	 }
       x = z;   
     
       if(f == 0.0)
	 {
	   printf("FATAL ERROR in LnGamma\n");
	   exit(-1);
	 }
       f=-log(f);
     }
   
   z = 1/(x*x);
   
   result = f + (x-0.5)*log(x) - x + .918938533204673 
	  + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	       +.083333333333333)/x;  

   return result;
}



double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
	   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-8, overflow=1e30;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];


   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   
   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term*=x/rn;   gin+=term;

   if (term > accurate) goto l20;
   gin*=factor/p;
   goto l50;
 l30:  
   /* (2) continued fraction */
   a=1-p;   b=a+x+1;  term=0;
   pn[0]=1;  pn[1]=x;  pn[2]=x+1;  pn[3]=x*b;
   gin=pn[2]/pn[3];   
 l32:  
   a++;  
   b+=2;  
   term++;   
   an=a*term;
   for (i=0; i<2; i++) 
     pn[i+4]=b*pn[i+2]-an*pn[i];
   if (pn[5] == 0) goto l35;
   rn=pn[4]/pn[5];   
   dif=fabs(gin-rn);  
   if (dif>accurate) goto l34;
   if (dif<=accurate*rn) goto l42;
 l34:   
   gin=rn;
 l35: 
   for (i=0; i<4; i++) 
     pn[i]=pn[i+2];
   if (fabs(pn[4]) < overflow)            
     goto l32;        
   
   for (i=0; i<4; i++) 
     pn[i]/=overflow;

   
   goto l32;
 l42:  
   gin=1-factor*gin;

 l50: 
   return (gin);
}




double PointNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.

*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) return (-9999);

   y = sqrt (log(1/(p1*p1)));   
   z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   return (p<0.5 ? -z : z);
}


double PointChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;
  
   if (p<.000002 || p>.999998 || v<=0) return (-1);
  
   g = LnGamma(v/2);
   
   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3:    
   x=PointNormal (p);
   p1=0.222222/v;   ch=v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)  ch=-2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;   
   if ((t=IncompleteGamma (p1, xx, g))< 0.0) 
     {
       printf ("IncompleteGamma \n");      
       return (-1);
     }
  
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}


void makeGammaCats(tree *tr, int model)
{
  int i,
    K = tr->gammaCategories;
  double gap05, t, factor, lnga1, alfa, beta;
  double
    *gammaProbs = (double *)malloc(K * sizeof(double));

  alfa = beta = tr->alphas[model];

  if(alfa < ALPHA_MIN)
    {
      printf("GAMMA_CATS alfa smaller ALPHA_MIN\n");
      exit(-1);
    }

  gap05 = 1.0 / (2.0 * K);
  factor = alfa / beta * K;
  
  if(K==1)
     {
       printf("ERROR in makeGammaCats\n");
       exit(-1);
     }

   lnga1=LnGamma(alfa+1);
   
   for (i=0; i<K-1; i++)
     gammaProbs[i]=PointGamma((i+1.0)/K, alfa, beta);
  
   for (i=0; i<K-1; i++)
     gammaProbs[i]=IncompleteGamma(gammaProbs[i]*beta, alfa+1, lnga1);

   tr->gammaRates[model * K] = gammaProbs[0]*factor;

   tr->gammaRates[model * K + (K-1)] = (1 - gammaProbs[K-2])*factor;

   for (i=1; i<K-1; i++){
     tr->gammaRates[model * K + i] = (gammaProbs[i] - gammaProbs[i-1])*factor;
   }

   free(gammaProbs);

   return;
}


void initModel(tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef)
{  
  int i, j, models;
  double  temp, wtemp;
  int model;

  optimizeRatesInvocations = 1;  
  optimizeRateCategoryInvocations = 1;  
  optimizeAlphaInvocations = 1;   
  optimizeTTRatioInvocations = 1;

  models = tr->NumberOfModels + tr->numberOfUserModels;

  for(model = 0; model < models; model++)
    {    
      for(i = 0; i < tr->numberOfRates; i++)	    
	tr->initialRates[model * tr->numberOfRates  + i] = 0.5;	       	    

      tr->userModelsWeights[model] = 1.0/models;
    }

  switch(adef->model)
    {       
    case M_PROTGAMMA:      
      if(adef->useMultipleModel) 
	{
	  newview = newviewGTRGAMMAPROTMULT;
	  makenewz = makenewzGTRGAMMAPROTMULT;
	  evaluate = evaluateGTRGAMMAPROTMULT;

	  evaluatePartition = evaluateGTRGAMMAPROTMULTPARTITION;
	  newviewPartition = newviewGTRGAMMAPROTMULTPARTITION;	
	}
      else if (!adef->useMultipleUserModels && !adef->useCFmodel && !adef->useCF9model && !adef->useCF15model && !adef->useCAT_C20model && !adef->useCAT_C20Fmodel)  /* not cF or cF-like mixture models */
	{
	  newview = newviewGTRGAMMAPROT;
	  makenewz = makenewzGTRGAMMAPROT;
	  evaluate = evaluateGTRGAMMAPROT;	  	 
	}
      else {       
	  newview = myNewviewGTRGAMMAPROTmix;
	  makenewz = myMakenewzGTRGAMMAPROTmix;
	  evaluate = myEvaluateGTRGAMMAPROTmix;	  	 
      }
    
      for(j = 0; j < tr->cdta->endsite; j++)
	tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
      
      baseFrequenciesGTR(rdta, cdta, tr, adef);                 
      
      /* only one alpha is used for multiple user models */
      if (adef->useMultipleUserModels || adef->useCFmodel && !adef->useCF9model || adef->useCF15model || adef->useCAT_C20model || !adef->useMultipleModel) {
	tr->alphas[0] = 1.0;
	makeGammaCats(tr, 0);
      }
      else {
	for (model = 0; model < tr->NumberOfModels; model ++) {
	  tr->alphas[model] = 1.0;
	  makeGammaCats(tr, model);
	}
      }
	  
      for(model = 0; model < models; model++)
	initReversibleGTR(rdta, cdta, tr, adef, model); /* but the rate matrices are still initialized for each model separately */
	  
      break;           
    case M_PROTCAT:   
      if(adef->useMultipleModel)
	{
	  newview = newviewGTRCATPROTMULT; 
	  makenewz = makenewzGTRCATPROTMULT; 
	  evaluate = evaluateGTRCATPROTMULT;	 
	  
	  evaluatePartial = evaluatePartialGTRCATPROTMULT;
	}
      else
	{
	  newview = newviewGTRCATPROT;
	  makenewz = makenewzGTRCATPROT;
	  evaluate = evaluateGTRCATPROT;
      
	  evaluatePartial = evaluatePartialGTRCATPROT;
	}
      
      for(j = 0; j < tr->cdta->endsite; j++)
	tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
      for (j = 0; j < tr->cdta->endsite; j++) 
	{
	  tr->cdta->patrat[j] = temp = 1.0;
	  tr->cdta->patratStored[j] = 1.0;
	  tr->cdta->rateCategory[j] = 0;
	  tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
	  tr->cdta->wr2[j] = temp * wtemp;
	}     
      
      baseFrequenciesGTR(rdta, cdta, tr, adef);

      tr->NumberOfCategories = 1;   
      for(model = 0; model < tr->NumberOfModels; model++)		
	initReversibleGTR(rdta, cdta, tr, adef, model);	  
	
	         
      break; 
    case M_GTRGAMMA:
      if(adef->useMultipleModel)
	{
	  

	  newview = newviewGTRGAMMAMULT;
	  makenewz = makenewzGTRGAMMAMULT;
	  evaluate = evaluateGTRGAMMAMULT;

	  evaluatePartition = evaluateGTRGAMMAMULTPARTITION;
	  newviewPartition = newviewGTRGAMMAMULTPARTITION;
	}
      else
	{	  
	  newview  = newviewGTRGAMMA;
	  makenewz = makenewzGTRGAMMA;
	  evaluate = evaluateGTRGAMMA;	  	 		           
	}  
      for(j = 0; j < tr->cdta->endsite; j++)
	tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
      baseFrequenciesGTR(rdta, cdta, tr, adef);                 
      for(model = 0; model < tr->NumberOfModels; model++)
	{	
	  tr->alphas[model] = 1.0;
	  initReversibleGTR(rdta, cdta, tr, adef, model);     
	  makeGammaCats(tr, model);      
	}
      
      break;
    case M_GTRCAT:
      if(adef->useMultipleModel)
	{	
	  newview           = newviewGTRCATMULT; 
	  makenewz          = makenewzGTRCATMULT; 
	  evaluate          = evaluateGTRCATMULT;

	  evaluatePartial   = evaluatePartialGTRCATMULT;
	  
	  evaluatePartition = evaluateGTRCATMULTPARTITION;
	  newviewPartition  = newviewPARTITIONGTRCATMULT;
	}
      else
	{      
	  newview         = newviewGTRCAT; 
	  makenewz        = makenewzGTRCAT;
	  evaluate        = evaluateGTRCAT;

	  evaluatePartial = evaluatePartialGTRCAT;	 
	}
   
      for (j = 0; j < tr->cdta->endsite; j++) 
	{
	  tr->cdta->patrat[j] = temp = 1.0;
	  tr->cdta->patratStored[j] = 1.0;
	  tr->cdta->rateCategory[j] = 0;
	  tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
	  tr->cdta->wr2[j] = temp * wtemp;
	}     
     
      baseFrequenciesGTR(rdta, cdta, tr,adef);           
      tr->NumberOfCategories = 1;    
      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  tr->alphas[model] = 1.0; /* GAMMA-APPROX */
	  makeGammaCats(tr, model); /*GAMMA-APPROX */
	  initReversibleGTR(rdta, cdta, tr, adef, model);            
	}
      break;    
    }
    
    
  if(adef->useMultipleModel)
    {
      tr->fracchange = 0;
      for(model = 0; model < models; model++)
	{
	  tr->fracchange += tr->fracchanges[model];
	}
      tr->fracchange /= ((double)models);
    }

  else tr->fracchange = 1.0;
}
